from __future__ import annotations

import asyncio

import enka


async def main() -> None:
    async with enka.HSRClient(enka.hsr.Language.TRADITIONAL_CHINESE) as api:
        # Update assets
        await api.update_assets()

        try:
            response = await api.fetch_showcase(809162009)
        except enka.errors.PlayerDoesNotExistError:
            return print("Player does not exist.")
        except enka.errors.GameMaintenanceError:
            return print("Game is in maintenance.")

        print("Name:", response.player.nickname)
        print("Level:", response.player.level)
        print("Equilbrium level:", response.player.equilibrium_level)
        print("Achievements:", response.player.stats.achievement_count)
        print("Light Cones:", response.player.stats.light_cone_count)
        print("Characters:", response.player.stats.character_count)
        print("Profile picture side icon:", response.player.icon)

        for character in response.characters:
            print("\n===============================\n")
            print(
                f"Lv. {character.level}/{character.max_level} {character.name} (E{character.eidolons_unlocked})"
            )
            print(f"Rarity: {character.rarity} ★")
            print("Element:", character.element.name.title())
            print("Path:", character.path.name.title())
            print("Round icon:", character.icon.round)

            lc = character.light_cone
            if lc is not None:
                print("\nLight Cone:")
                print(f"Lv. {lc.level}/{lc.max_level} {lc.name} (S{lc.superimpose})")
                print(f"Rarity: {lc.rarity} ★")
                for stat in lc.stats:
                    print(stat.name, stat.formatted_value)

            print("\nStats:")
            for stat in character.stats.values():
                if stat.value == 0 or stat.type.value in enka.hsr.DMG_BONUS_PROPS.values():
                    continue
                print(stat.name, stat.formatted_value)

            dmg_bonus = character.highest_dmg_bonus_stat
            print(dmg_bonus.name, dmg_bonus.formatted_value)

            print("\nRelics:")
            for relic in character.relics:
                main_stat = relic.main_stat
                print(
                    f"+{relic.level} {relic.set_name}: {main_stat.name} {main_stat.formatted_value}"
                )
                for substat in relic.sub_stats:
                    print(f"- {substat.name} {substat.formatted_value}")
                print()
    return None


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any, Final, Literal, overload

from loguru import logger

from ..assets.data import TextMap
from ..assets.hsr.manager import HSR_ASSETS
from ..calc.hsr import LayerGenerator, PropState
from ..enums.enum import Game
from ..enums.hsr import Element, Language, Path, StatType, TraceType
from ..models.hsr import CharacterIcon, LightConeIcon, Player, ShowcaseResponse, Stat
from ..models.hsr.build import Build
from ..models.hsr.character import Eidolon
from .base import BaseClient

if TYPE_CHECKING:
    from ..models.enka.owner import Owner
    from ..models.hsr.character import Character, LightCone, Relic, Trace
    from .cache import BaseTTLCache

__all__ = ("HSRClient",)

API_URL: Final[str] = "https://enka.network/api/hsr/uid/{uid}"


class HSRClient(BaseClient):
    """The main client to interact with the Enka Network Honkai Star Rail API.

    Args:
        lang (Language | str): The language to use for the client, defaults to Language.ENGLISH.
        headers (dict[str, Any] | None): The headers to use for the client, defaults to None.
        cache (BaseTTLCache | None): The cache to use for the client, defaults to None.
        use_enka_icons (bool): Whether to get stat icons from Enka, defaults to True.
    """

    def __init__(
        self,
        lang: Language | str = Language.ENGLISH,
        *,
        headers: dict[str, Any] | None = None,
        cache: BaseTTLCache | None = None,
        use_enka_icons: bool = True,
    ) -> None:
        super().__init__(Game.HSR, headers=headers, cache=cache)

        self._lang = self._convert_lang(lang)
        self._use_enka_icons = use_enka_icons
        self._assets = HSR_ASSETS

    def _convert_lang(self, lang: Language | str) -> Language:
        if not isinstance(lang, Language):
            try:
                lang = Language(lang)
            except ValueError as e:
                available_langs = ", ".join(lang.value for lang in Language)
                msg = f"Invalid language: {lang}, must be one of {available_langs}"
                raise ValueError(msg) from e

        return lang

    @property
    def lang(self) -> Language:
        return self._lang

    @lang.setter
    def lang(self, lang: Language | str) -> None:
        self._lang = self._convert_lang(lang)
        self._text_map = TextMap(self._lang, self._assets.text_map)

    async def __aenter__(self) -> HSRClient:
        await self.start()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return await super().__aexit__(exc_type, exc_val, exc_tb)

    def _post_process_relic(self, relic: Relic) -> None:
        text_map = self._text_map
        relic.set_name = text_map[relic.set_name]

        relic_data = self._assets.relic_data[str(relic.id)]
        relic.icon = self._get_icon(relic_data["Icon"])
        relic.rarity = relic_data["Rarity"]

        for stat in relic.stats:
            stat.name = text_map[stat.type.value]
            stat.icon = self._get_icon(
                self._assets.property_config_data[stat.type.value], enka=self._use_enka_icons
            )

    def _post_process_light_cone(self, light_cone: LightCone) -> None:
        text_map = self._text_map
        data = self._assets.light_cones_data[str(light_cone.id)]

        light_cone.name = text_map[light_cone.name]
        light_cone.rarity = data["Rarity"]
        light_cone.icon = LightConeIcon(light_cone_id=light_cone.id)
        light_cone.path = Path(data["AvatarBaseType"])

        for stat in light_cone.stats:
            stat.name = text_map[stat.type.value]
            stat.icon = self._get_icon(
                self._assets.property_config_data[stat.type.value], enka=self._use_enka_icons
            )

    def _post_process_trace(self, trace: Trace, unlocked_eidolon_ids: list[int]) -> None:
        skill_tree_data = self._assets.skill_tree_data
        trace_data = skill_tree_data[str(trace.id)]

        try:
            skill_ids: list[int] = trace_data["skillIds"]
        except KeyError as e:
            msg = "Skill IDs not found in trace data, please update the assets with `update_assets`"
            raise RuntimeError(msg) from e

        trace.anchor = trace_data["anchor"]
        trace.icon = self._get_icon(
            trace_data["icon"], enka=self._use_enka_icons or "SkillIcons" in trace_data["icon"]
        )
        trace.type = TraceType(trace_data["pointType"])
        trace.max_level = trace_data["maxLevel"]

        for eidolon_id in unlocked_eidolon_ids:
            eidolon_data = self._assets.eidolon_data[str(eidolon_id)]

            for skill_id in skill_ids:
                if str(skill_id) in eidolon_data["SkillAddLevelList"]:
                    trace.level += eidolon_data["SkillAddLevelList"][str(skill_id)]
                    trace.boosted = True
                    break

    def _post_process_character(self, character: Character) -> None:
        character.icon = CharacterIcon(character_id=character.id)

        character_data = self._assets.character_data[str(character.id)]
        text_map_hash = character_data["AvatarName"]["Hash"]
        character.name = self._text_map[text_map_hash]
        character.rarity = character_data["Rarity"]
        character.element = Element(character_data["Element"])
        character.path = Path(character_data["AvatarBaseType"])

        # Eidolons
        eidolon_ids: list[int] = character_data["RankIDList"]
        for i, eidolon_id in enumerate(eidolon_ids, start=1):
            eidolon_data = self._assets.eidolon_data[str(eidolon_id)]
            character.eidolons.append(
                Eidolon(
                    id=eidolon_id,
                    icon=self._get_icon(eidolon_data["IconPath"]),
                    unlocked=i <= character.eidolons_unlocked,
                )
            )

        for trace in character.traces:
            self._post_process_trace(
                trace, [e.id for e in character.eidolons[: character.eidolons_unlocked]]
            )
        for relic in character.relics:
            self._post_process_relic(relic)
        if character.light_cone is not None:
            self._post_process_light_cone(character.light_cone)

        # Credit to Enka Network for the following code
        generator = LayerGenerator(self._assets)
        prop_state = PropState()

        prop_state.add(generator.character(character))
        if character.light_cone is not None:
            prop_state.add(generator.light_cone(character.light_cone))

            if character.light_cone.path == character.path:
                prop_state.add(generator.light_cone_skill(character.light_cone))

        prop_state.add(generator.relics(character.relics))
        prop_state.add(generator.relic_set(character.relics))
        prop_state.add(generator.skill_tree(character))

        props = prop_state.sum()

        final_stats: dict[StatType, float] = {
            StatType.MAX_HP: props.hp,
            StatType.ATK: props.attack,
            StatType.DEF: props.defence,
            StatType.SPD: props.speed,
            StatType.CRIT_RATE: props.critical_chance,
            StatType.CRIT_DMG: props.critical_damage,
            StatType.BREAK_EFFECT: props.break_damage,
            StatType.HEALING_BOOST: props.heal_ratio,
            StatType.ENERGY_REGEN_RATE: props.sp_ratio,
            StatType.EFFECT_HIT_RATE: props.status_probability,
            StatType.EFFECT_RES: props.status_resistance,
            StatType.PHYSICAL_DMG_BOOST: props.physical_damage,
            StatType.FIRE_DMG_BOOST: props.fire_damage,
            StatType.ICE_DMG_BOOST: props.ice_damage,
            StatType.LIGHTNING_DMG_BOOST: props.thunder_damage,
            StatType.WIND_DMG_BOOST: props.wind_damage,
            StatType.QUANTUM_DMG_BOOST: props.quantum_damage,
            StatType.IMAGINARY_DMG_BOOST: props.imaginary_damage,
        }

        character.stats = {
            stat_type: Stat(
                type=stat_type,
                value=value,
                name=self._text_map[stat_type.value],
                icon=self._get_icon(
                    self._assets.property_config_data[stat_type.value], enka=self._use_enka_icons
                ),
            )
            for stat_type, value in final_stats.items()
        }

    def _post_process_player(self, player: Player) -> None:
        player.icon = self._get_icon(self._assets.avatar_data[player.icon]["Icon"])

    def _post_process_showcase(self, showcase: ShowcaseResponse) -> None:
        for character in showcase.characters:
            self._post_process_character(character)
        self._post_process_player(showcase.player)

    def _get_icon(self, icon: str, *, enka: bool = True) -> str:
        icon = icon.replace(".png", "")
        if enka:
            return f"https://enka.network/ui/hsr/{icon}.png"
        return f"https://raw.githubusercontent.com/Mar-7th/StarRailRes/master/icon/property/{icon.split('/')[-1]}.png"

    async def start(self) -> None:
        await super().start()
        await self._assets.load(self.session)
        self._text_map = TextMap(self.lang, self._assets.text_map)

    async def update_assets(self) -> None:
        """Update game assets."""
        logger.info("Updating HSR assets")
        await self._assets.update(self.session)
        self._text_map = TextMap(self.lang, self._assets.text_map)
        logger.info("HSR assets updated")

    @overload
    async def fetch_showcase(
        self, uid: str | int, *, raw: Literal[False] = False
    ) -> ShowcaseResponse: ...
    @overload
    async def fetch_showcase(
        self, uid: str | int, *, raw: Literal[True] = True
    ) -> dict[str, Any]: ...
    async def fetch_showcase(
        self, uid: str | int, *, raw: bool = False
    ) -> ShowcaseResponse | dict[str, Any]:
        """Fetch the player showcase of the given UID.

        Args:
            uid: The UID of the user.
            raw: Whether to return the raw data, defaults to False.

        Returns:
            The parsed or raw showcase data.
        """
        url = API_URL.format(uid=uid)

        data = await self._request(url)
        if raw:
            return data

        data = copy.deepcopy(data)
        showcase = ShowcaseResponse(**data)
        self._post_process_showcase(showcase)
        return showcase

    def parse_showcase(self, data: dict[str, Any]) -> ShowcaseResponse:
        """Parse the given showcase data.

        Args:
            data: The showcase data.

        Returns:
            The parsed showcase response.
        """
        data = copy.deepcopy(data)
        showcase = ShowcaseResponse(**data)
        self._post_process_showcase(showcase)
        return showcase

    async def fetch_builds(self, owner: Owner) -> dict[str, list[Build]]:
        """Fetch the character builds of the given owner.

        Args:
            owner: The owner of the builds.

        Returns:
            Character ID to list of builds mapping.
        """
        url = f"https://enka.network/api/profile/{owner.username}/hoyos/{owner.hash}/builds/"
        data = await self._request(url)
        result: dict[str, list[Build]] = {}

        for key, builds in data.items():
            result[key] = []
            for build in builds:
                build_ = Build(**build)
                self._post_process_character(build_.character)
                result[key].append(build_)

        return result



from __future__ import annotations

from pydantic import BaseModel, Field, field_validator

__all__ = ("Player", "PlayerStats")


class PlayerStats(BaseModel):
    """Represents the statistics of a HSR player.

    Attributes:
        achievement_count: The number of achievements the player has.
        light_cone_count: The number of light cones the player has.
        character_count: The number of characters the player has.
        max_simulated_universe_world: The maximum world completed in the simulated universe.
        book_count: The number of books the player has, added in game version 2.2
        relic_count: The number of relics the player has, added in game version 2.2
        music_count: The number of music tracks the player has, added in game version 2.2
    """

    achievement_count: int = Field(0, alias="achievementCount")
    light_cone_count: int = Field(0, alias="equipmentCount")
    character_count: int = Field(alias="avatarCount")
    max_simulated_universe_world: int | None = Field(None, alias="maxRogueChallengeScore")

    # New after game version 2.2
    book_count: int | None = Field(None, alias="bookCount")
    relic_count: int | None = Field(None, alias="relicCount")
    music_count: int | None = Field(None, alias="musicCount")


class Player(BaseModel):
    """Represents a HSR player.

    Attributes:
        nickname: The player's nickname.
        signature: The player's signature.
        uid: The player's UID.
        level: The player's level.
        equilibrium_level: The player's equilibrium level.
        friend_count: The number of friends the player has.
        stats: The player's statistics.
    """

    nickname: str
    signature: str = ""
    uid: int
    level: int
    equilibrium_level: int = Field(0, alias="worldLevel")
    friend_count: int = Field(0, alias="friendCount")
    stats: PlayerStats = Field(alias="recordInfo")

    # The following fields are added in post-processing
    icon: str = Field("", alias="headIcon")

    @field_validator("icon", mode="before")
    @classmethod
    def _stringify_icon(cls, value: int) -> str:
        return str(value)



from __future__ import annotations

from ..data import AssetData
from ..manager import AssetManager
from .file_paths import (
    AVATAR_DATA_PATH,
    CHARACTER_DATA_PATH,
    EIDOLON_DATA_PATH,
    LIGHT_CONE_DATA_PATH,
    META_DATA_PATH,
    PROPERTY_CONFIG_PATH,
    RELIC_DATA_PATH,
    SKILL_TREE_DATA_PATH,
    TEXT_MAP_PATH,
)


class HSRAssetManager(AssetManager):
    """Honkai Star Rail asset manager."""

    def __init__(self) -> None:
        self.text_map = AssetData(TEXT_MAP_PATH)
        self.character_data = AssetData(CHARACTER_DATA_PATH)
        self.skill_tree_data = AssetData(SKILL_TREE_DATA_PATH)
        self.light_cones_data = AssetData(LIGHT_CONE_DATA_PATH)
        self.relic_data = AssetData(RELIC_DATA_PATH)
        self.meta_data = AssetData(META_DATA_PATH)
        self.avatar_data = AssetData(AVATAR_DATA_PATH)
        self.property_config_data = AssetData(PROPERTY_CONFIG_PATH)
        self.eidolon_data = AssetData(EIDOLON_DATA_PATH)

        self._assets = (
            self.text_map,
            self.character_data,
            self.skill_tree_data,
            self.light_cones_data,
            self.relic_data,
            self.meta_data,
            self.avatar_data,
            self.property_config_data,
            self.eidolon_data,
        )


HSR_ASSETS = HSRAssetManager()



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, computed_field

from .icon import Icon

__all__ = ("Costume",)


class Costume(BaseModel):
    """Represents a character's costume.

    Attributes:
       id: The costume's ID.
       icon: The costume's icon.
    """

    id: int
    data: dict[str, Any]

    @computed_field
    @property
    def icon(self) -> Icon:
        """The costume's icon."""
        return Icon(side_icon_ui_path=self.data["sideIconName"], is_costume=True)



from __future__ import annotations

import copy
import math
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Final, Literal, overload

from loguru import logger

from ..assets.data import TextMap
from ..assets.zzz.manager import ZZZ_ASSETS
from ..enums import zzz as enums
from ..enums.enum import Game
from ..models import zzz as models
from .base import BaseClient

if TYPE_CHECKING:
    from .cache import BaseTTLCache

__all__ = ("ZZZClient",)

API_URL: Final[str] = "https://enka.network/api/zzz/uid/{uid}"


class ZZZClient(BaseClient):
    """The main client to interact with the Enka Network Zenless Zone Zero API.

    Args:
        lang (Language | str): The language to use for the client, defaults to Language.ENGLISH.
        headers (dict[str, Any] | None): The headers to use for the client, defaults to None.
        cache (BaseTTLCache | None): The cache to use for the client, defaults to None.
    """

    def __init__(
        self,
        lang: enums.Language | str = enums.Language.ENGLISH,
        *,
        headers: dict[str, Any] | None = None,
        cache: BaseTTLCache | None = None,
    ) -> None:
        super().__init__(Game.ZZZ, headers=headers, cache=cache)

        self._lang = self._convert_lang(lang)
        self._assets = ZZZ_ASSETS

    def _convert_lang(self, lang: enums.Language | str) -> enums.Language:
        if isinstance(lang, str):
            try:
                lang = enums.Language(lang)
            except ValueError as e:
                available_langs = ", ".join(lang.value for lang in enums.Language)
                msg = f"Invalid language: {lang}, must be one of {available_langs}"
                raise ValueError(msg) from e

        return lang

    @property
    def lang(self) -> enums.Language:
        return self._lang

    @lang.setter
    def lang(self, lang: enums.Language | str) -> None:
        self._lang = self._convert_lang(lang)
        self._text_map = TextMap(self._lang, self._assets.text_map)

    async def __aenter__(self) -> ZZZClient:
        await self.start()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return await super().__aexit__(exc_type, exc_val, exc_tb)

    async def start(self) -> None:
        await super().start()
        await self._assets.load(self.session)
        self._text_map = TextMap(self.lang, self._assets.text_map)

    async def update_assets(self) -> None:
        """Update game assets."""
        logger.info("Updating ZZZ assets")
        await self._assets.update(self.session)
        self._text_map = TextMap(self.lang, self._assets.text_map)
        logger.info("ZZZ assets updated")

    def _calc_disc_stats(self, disc: models.DriveDisc) -> None:
        # See https://api.enka.network/#/docs/zzz/api?id=drive-disc for the formula
        disc_level_data = next(
            (
                d
                for d in self._assets.equipment_level["Items"]
                if d["Level"] == disc.level and d["Rarity"] == disc.rarity_num
            ),
            None,
        )
        if disc_level_data is None:
            logger.warning(
                f"Drive disc level data not found for level={disc.level} and rarity={disc.rarity_num}"
            )
            return

        main_stat_val = disc.main_stat.value * (1 + disc_level_data["EnhanceRate"] / 10000)
        disc.main_stat.value = math.floor(main_stat_val)

        main_stat_data = self._assets.property[str(disc.main_stat.type.value)]
        disc.main_stat.name = self._text_map[main_stat_data["Name"]]
        disc.main_stat.format = main_stat_data["Format"]

        for sub in disc.sub_stats:
            sub.value *= sub.roll_times

            sub_data = self._assets.property[str(sub.type.value)]
            sub.name = self._text_map[sub_data["Name"]]
            sub.format = sub_data["Format"]

    def _get_agent_base_stats(self, agent: models.Agent) -> defaultdict[int, int]:
        stats: defaultdict[int, int] = defaultdict(int)
        data = self._assets.avatars[str(agent.id)]
        base_props: dict[str, int] = data["BaseProps"]
        growth_props: dict[str, int] = data["GrowthProps"]
        promotion_props: list[dict[str, int]] = data["PromotionProps"]
        core_enhancement_props: list[dict[str, int]] = data["CoreEnhancementProps"]

        for prop_id_str, prop_val in base_props.items():
            prop_id = int(prop_id_str)
            growth_val = (growth_props.get(prop_id_str, 0) * (agent.level - 1)) / 10000
            promotion_val = promotion_props[agent.promotion - 1].get(prop_id_str, 0)
            core_enhancement_val = core_enhancement_props[agent.core_skill_level_num].get(
                prop_id_str, 0
            )
            base_total_val = (
                prop_val + math.floor(growth_val) + promotion_val + core_enhancement_val
            )
            stats[prop_id] += base_total_val

        return stats

    def _add_engine_stats(self, stats: defaultdict[int, int], engine: models.WEngine) -> None:
        stats[engine.main_stat.type.value] += engine.main_stat.value
        stats[engine.sub_stat.type.value] += engine.sub_stat.value

    def _add_disc_stats(self, stats: defaultdict[int, int], discs: list[models.DriveDisc]) -> None:
        for disc in discs:
            stats[disc.main_stat.type.value] += disc.main_stat.value
            for sub in disc.sub_stats:
                stats[sub.type.value] += sub.value

    def _add_disc_set_bonus_stats(
        self, stats: defaultdict[int, int], discs: list[models.DriveDisc]
    ) -> None:
        sets: defaultdict[int, int] = defaultdict(int)
        for disc in discs:
            sets[disc.set_id] += 1

        for set_id, count in sets.items():
            if count < 2:
                continue

            set_data = self._assets.equipments["Suits"].get(str(set_id))
            if set_data is None:
                logger.warning(f"Set data not found for {set_id=}")
                continue

            set_effect: dict[str, int] = set_data["SetBonusProps"]
            for prop_id_str, prop_val in set_effect.items():
                stats[int(prop_id_str)] += prop_val

    def _calculate_final_stats(self, raw_stats: defaultdict[int, int]) -> dict[int, int]:
        final_stats: dict[int, int] = {
            stat_type.value: 0
            for stat_type in enums.StatType
            if str(stat_type.value).endswith("01")  # Initialize base stats
        }

        # Apply base stats first
        for prop_id, prop_val in raw_stats.items():
            if str(prop_id).endswith("01"):
                if prop_id not in final_stats:
                    final_stats[prop_id] = 0
                final_stats[prop_id] += prop_val

        # Apply percentage stats
        for prop_id, prop_val in raw_stats.items():
            if str(prop_id).endswith("02"):
                base_prop_id = prop_id - 1
                base_stat = final_stats.get(base_prop_id, 0)
                stat = base_stat * (1 + prop_val / 10000)
                # Max HP is rounded up, others down
                rounded = math.ceil(stat) if base_prop_id == 11101 else math.floor(stat)
                final_stats[base_prop_id] = rounded

        # Apply flat stats
        for prop_id, prop_val in raw_stats.items():
            if str(prop_id).endswith("03"):
                base_prop_id = prop_id - 2
                base_stat = final_stats.get(base_prop_id, 0)
                final_stats[base_prop_id] += prop_val

        return final_stats

    def _apply_special_agent_passives(
        self, final_stats: dict[int, int], agent: models.Agent
    ) -> None:
        if agent.id == 1121:  # Ben
            # Ben's initial ATK increases along with his initial DEF. He gains X% of his initial DEF as ATK.
            scalings = (0.4, 0.46, 0.52, 0.6, 0.66, 0.72, 0.8)
            def_stat_id = 13101
            atk_stat_id = 11101

            if def_stat_id in final_stats and atk_stat_id in final_stats:
                final_stats[atk_stat_id] += math.floor(
                    final_stats[def_stat_id] * scalings[agent.core_skill_level_num]
                )

    def _assign_final_stats_to_agent(
        self, final_stats: dict[int, int], agent: models.Agent
    ) -> None:
        for prop_id, prop_val in final_stats.items():
            stat_type = enums.StatType(prop_id)
            prop_data = self._assets.property[str(prop_id)]
            agent.stats[stat_type] = models.Stat(
                type=stat_type,
                value=prop_val,
                name=self._text_map[prop_data["Name"]],
                format=prop_data["Format"],
            )

    def _calc_agent_stats(self, agent: models.Agent) -> None:
        # See https://api.enka.network/#/docs/zzz/api?id=agent-stats for the formula

        # 1. Get base stats from agent level, promotion, core skills
        raw_stats = self._get_agent_base_stats(agent)

        # 2. Add engine stats
        if agent.w_engine is not None:
            self._add_engine_stats(raw_stats, agent.w_engine)

        # 3. Add disc stats
        self._add_disc_stats(raw_stats, agent.discs)

        # 4. Add disc set bonus stats
        self._add_disc_set_bonus_stats(raw_stats, agent.discs)

        # 5. Calculate final stats (applying %, flat bonuses)
        final_stats = self._calculate_final_stats(raw_stats)

        # 6. Apply special agent passives
        self._apply_special_agent_passives(final_stats, agent)

        # 7. Assign final stats to agent object
        self._assign_final_stats_to_agent(final_stats, agent)

    def _calc_engine_stats(self, engine: models.WEngine) -> None:
        # See https://api.enka.network/#/docs/zzz/api?id=w-engine for the formula

        level_data = next(
            (
                d
                for d in self._assets.weapon_level["Items"]
                if d["Level"] == engine.level and d["Rarity"] == engine.rarity_num
            ),
            None,
        )
        if level_data is None:
            logger.warning(
                f"W-Engine level data not found for level={engine.level} and rarity={engine.rarity_num}"
            )
            return

        enhance_rate = level_data["EnhanceRate"]

        star_data = next(
            (
                d
                for d in self._assets.weapon_star["Items"]
                if d["Rarity"] == engine.rarity_num and d["BreakLevel"] == engine.modification
            ),
            None,
        )
        if star_data is None:
            logger.warning(f"W-Engine star data not found for rarity={engine.rarity_num}")
            return

        star_rate = star_data["StarRate"]
        rand_rate = star_data["RandRate"]

        main_stat_val = engine.main_stat.value * (1 + enhance_rate / 10000 + star_rate / 10000)
        engine.main_stat.value = math.floor(main_stat_val)

        sub_stat_val = engine.sub_stat.value * (1 + rand_rate / 10000)
        engine.sub_stat.value = math.floor(sub_stat_val)

    def _post_process_engine(self, engine: models.WEngine) -> None:
        engine_data = self._assets.weapons[str(engine.id)]
        engine.name = self._text_map[engine_data["ItemName"]]
        engine.rarity_num = engine_data["Rarity"]
        engine.specialty = enums.ProfessionType(engine_data["ProfessionType"])
        engine.icon = f"https://enka.network{engine_data['ImagePath']}"

        main_stat = engine_data["MainStat"]
        main_stat_data = self._assets.property[str(main_stat["PropertyId"])]
        engine.main_stat = models.Stat(
            type=enums.StatType(main_stat["PropertyId"]),
            value=main_stat["PropertyValue"],
            name=self._text_map[main_stat_data["Name"]],
            format=main_stat_data["Format"],
        )

        sub_stat = engine_data["SecondaryStat"]
        sub_stat_data = self._assets.property[str(sub_stat["PropertyId"])]
        engine.sub_stat = models.Stat(
            type=enums.StatType(sub_stat["PropertyId"]),
            value=sub_stat["PropertyValue"],
            name=self._text_map[sub_stat_data["Name"]],
            format=sub_stat_data["Format"],
        )

        self._calc_engine_stats(engine)

    def _post_process_disc(self, disc: models.DriveDisc) -> None:
        disc_data = self._assets.equipments["Items"][str(disc.id)]
        disc.rarity_num = disc_data["Rarity"]
        disc.set_id = disc_data["SuitId"]
        self._calc_disc_stats(disc)

    def _post_process_agent(self, agent: models.Agent) -> None:
        data = self._assets.avatars[str(agent.id)]
        agent.name = self._text_map[data["Name"]]
        agent.rarity_num = data["Rarity"]
        agent.elements = [enums.Element(e) for e in data["ElementTypes"]]
        agent.icon = models.AgentIcon(
            filename=data["Image"].removeprefix("/ui/zzz/").removesuffix(".png")
        )
        agent.sig_engine_id = data["WeaponId"]
        agent.color = models.AgentColor(**data["Colors"])
        agent.highlight_stats = [enums.StatType(stat) for stat in data["HighlightProps"]]

        for disc in agent.discs:
            self._post_process_disc(disc)

        if agent.w_engine is not None:
            self._post_process_engine(agent.w_engine)
        self._calc_agent_stats(agent)

    def _post_process_title(self, title: models.Title) -> None:
        title_data = self._assets.titles[str(title.id)]
        title.text = self._text_map[title_data["TitleText"]]
        title.color1 = f"#{title_data['ColorA']}"
        title.color2 = f"#{title_data['ColorB']}"

    def _post_process_player(self, player: models.Player) -> None:
        self._post_process_title(player.title)

        # Namecard
        namecard_id = player.namecard_id
        player.namecard = models.Namecard(
            id=namecard_id, icon=self._assets.namecards[str(namecard_id)]["Icon"]
        )

    def _post_process_showcase(self, showcase: models.ShowcaseResponse) -> None:
        self._post_process_player(showcase.player)
        for agent in showcase.agents:
            self._post_process_agent(agent)

    @overload
    async def fetch_showcase(
        self, uid: str | int, *, raw: Literal[False] = False
    ) -> models.ShowcaseResponse: ...
    @overload
    async def fetch_showcase(
        self, uid: str | int, *, raw: Literal[True] = True
    ) -> dict[str, Any]: ...
    async def fetch_showcase(
        self, uid: str | int, *, raw: bool = False
    ) -> models.ShowcaseResponse | dict[str, Any]:
        """Fetch the player showcase of the given UID.

        Args:
            uid: The UID of the user.
            raw: Whether to return the raw data, defaults to False.

        Returns:
            The parsed or raw showcase data.
        """
        url = API_URL.format(uid=uid)

        data = await self._request(url)
        if raw:
            return data

        data = copy.deepcopy(data)
        showcase = models.ShowcaseResponse(**data)
        self._post_process_showcase(showcase)
        return showcase

    def parse_showcase(self, data: dict[str, Any]) -> models.ShowcaseResponse:
        """Parse the given showcase data.

        Args:
            data: The showcase data.

        Returns:
            The parsed showcase response.
        """
        data = copy.deepcopy(data)
        showcase = models.ShowcaseResponse(**data)
        self._post_process_showcase(showcase)
        return showcase



from __future__ import annotations

import pytest

from enka.errors import PlayerDoesNotExistError
from enka.zzz import Language, ZZZClient


async def test_update_assets(zzz_client: ZZZClient) -> None:
    await zzz_client.update_assets()


async def test_fetch_showcase(zzz_client: ZZZClient) -> None:
    await zzz_client.fetch_showcase("1300025292")


@pytest.mark.parametrize("lang", list(Language))
async def test_langs(lang: Language) -> None:
    async with ZZZClient(lang) as api:
        await api.fetch_showcase("1300025292")


async def test_raw_and_parse(zzz_client: ZZZClient) -> None:
    raw = await zzz_client.fetch_showcase("1300025292", raw=True)
    zzz_client.parse_showcase(raw)


async def test_not_exist_player(zzz_client: ZZZClient) -> None:
    with pytest.raises(PlayerDoesNotExistError):
        await zzz_client.fetch_showcase("1000000000")



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field, model_validator

from ..enka.owner import Owner
from .character import Character
from .player import Player

__all__ = ("ShowcaseResponse",)


class ShowcaseResponse(BaseModel):
    """
    Represents a HSR showcase response.

    Attributes:
        characters: The characters in the showcase.
        player: The player.
        ttl: The time to live of the response.
        uid: The UID of the showcase.
        owner: The owner of the showcase's account, if any.
    """

    characters: list[Character] = Field(alias="avatarDetailList")
    player: Player = Field(alias="detailInfo")
    ttl: int = 0
    uid: str
    owner: Owner | None = None

    @model_validator(mode="before")
    @classmethod
    def _flatten_data(cls, values: dict[str, Any]) -> dict[str, Any]:
        values["avatarDetailList"] = values["detailInfo"].pop("avatarDetailList", [])
        return values



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Any

import aiofiles
import aiofiles.os
import orjson
from loguru import logger

from ..assets.gi.file_paths import PATH_TO_SOURCE as GI_PATH_TO_SOURCE
from ..assets.hsr.file_paths import PATH_TO_SOURCE as HSR_PATH_TO_SOURCE
from ..assets.zzz.file_paths import PATH_TO_SOURCE as ZZZ_PATH_TO_SOURCE
from ..errors import AssetDownloadError, AssetKeyError

if TYPE_CHECKING:
    import pathlib

    import aiohttp

    from ..enums import gi, hsr, zzz

PATH_TO_SOURCE = GI_PATH_TO_SOURCE | HSR_PATH_TO_SOURCE | ZZZ_PATH_TO_SOURCE


class BaseAssetData:
    """Base class for asset data, this functions like a dictionary."""

    def __init__(self, data: dict[str, Any] | None = None) -> None:
        self._data: dict[str, Any] | None = data

    @property
    def data(self) -> dict[str, Any]:
        if self._data is None:
            msg = f"{self.__class__.__name__} is not loaded"
            raise RuntimeError(msg)
        return self._data

    @data.setter
    def data(self, value: dict[str, Any]) -> None:
        self._data = value

    def __getitem__(self, key: str) -> Any:
        try:
            return self.data[str(key)]
        except KeyError as e:
            raise AssetKeyError(key, self.__class__) from e

    def __iter__(self) -> Any:
        return iter(self.data)

    def values(self) -> Any:
        return self.data.values()

    def items(self) -> Any:
        return self.data.items()

    def get(self, key: str, default: Any = None) -> Any:
        text = self.data.get(str(key))
        if text is None:
            return default

        return text


class AssetData(BaseAssetData):
    def __init__(self, path: pathlib.Path) -> None:
        super().__init__()
        self._path = path

    async def _open_json(self) -> dict[str, Any] | None:
        logger.debug(f"Opening {self._path}")
        with contextlib.suppress(FileNotFoundError):
            async with aiofiles.open(self._path, encoding="utf-8") as f:
                return orjson.loads(await f.read())
        return None

    async def _download_json(self, session: aiohttp.ClientSession) -> None:
        url = PATH_TO_SOURCE[self._path]
        logger.debug(f"Downloading from {url}")

        async with session.get(url) as resp:
            if resp.status != 200:
                raise AssetDownloadError(resp.status, url)

            text = await resp.text()

            await aiofiles.os.makedirs(self._path.parent, exist_ok=True)
            async with aiofiles.open(self._path, "w", encoding="utf-8") as f:
                await f.write(text)

            self._data = orjson.loads(text)

    async def update(self, session: aiohttp.ClientSession) -> None:
        await self._download_json(session)
        logger.debug(f"Updated {self._path}")

    async def load(self, session: aiohttp.ClientSession) -> None:
        """Load or download the asset data"""
        if self._data is not None:
            # Prevent loading asset again
            return

        self._data = await self._open_json()
        if self._data is None:
            await self._download_json(session)


class TextMap(BaseAssetData):
    """Text map asset data."""

    def __init__(
        self, lang: gi.Language | hsr.Language | zzz.Language, text_map: AssetData
    ) -> None:
        super().__init__(text_map[lang.value])



from __future__ import annotations

from ..data import AssetData
from ..manager import AssetManager
from .file_paths import (
    CHARACTER_DATA_PATH,
    CONSTS_DATA_PATH,
    NAMECARD_DATA_PATH,
    PFPS_DATA_PATH,
    TALENTS_DATA_PATH,
    TEXT_MAP_PATH,
)


class GIAssetManager(AssetManager):
    """Genshin Impact asset manager."""

    def __init__(self) -> None:
        self.text_map = AssetData(TEXT_MAP_PATH)
        self.character_data = AssetData(CHARACTER_DATA_PATH)
        self.namecard_data = AssetData(NAMECARD_DATA_PATH)
        self.consts_data = AssetData(CONSTS_DATA_PATH)
        self.talents_data = AssetData(TALENTS_DATA_PATH)
        self.pfps_data = AssetData(PFPS_DATA_PATH)

        self._assets = (
            self.text_map,
            self.character_data,
            self.namecard_data,
            self.consts_data,
            self.talents_data,
            self.pfps_data,
        )


GI_ASSETS = GIAssetManager()



from __future__ import annotations

__all__ = ("CharacterIcon", "LightConeIcon")


from pydantic import BaseModel, computed_field


class CharacterIcon(BaseModel):
    """Represents a HSR character icon.

    Attributes:
        character_id: The character's ID.
    """

    character_id: int

    @computed_field
    @property
    def round(self) -> str:
        """Character icon in round shape.

        e.g. https://enka.network/ui/hsr/SpriteOutput/AvatarRoundIcon/1001.png
        """
        return f"https://enka.network/ui/hsr/SpriteOutput/AvatarRoundIcon/{self.character_id}.png"

    @computed_field
    @property
    def gacha(self) -> str:
        """Character gacha splash art.

        e.g. https://enka.network/ui/hsr/SpriteOutput/AvatarDrawCard/1001.png
        """
        return f"https://enka.network/ui/hsr/SpriteOutput/AvatarDrawCard/{self.character_id}.png"

    @computed_field
    @property
    def card(self) -> str:
        """Character icon in card shape.

        Provided by Project Yatta.
        e.g. https://api.yatta.top/hsr/assets/UI//avatar/medium/1001.png
        """
        return f"https://api.yatta.top/hsr/assets/UI//avatar/medium/{self.character_id}.png"


class LightConeIcon(BaseModel):
    """Represents a HSR light cone icon.

    Attributes:
        light_cone_id: The light cone's ID.
    """

    light_cone_id: int

    @computed_field
    @property
    def image(self) -> str:
        """Light cone icon image.

        e.g. https://enka.network/ui/hsr/SpriteOutput/LightConeFigures/20000.png
        """
        return f"https://enka.network/ui/hsr/SpriteOutput/LightConeFigures/{self.light_cone_id}.png"

    @computed_field
    @property
    def item(self) -> str:
        """Light cone icon item.

        Provided by Project Yatta.
        e.g. https://api.yatta.top/hsr/assets/UI//equipment/medium/20000.png
        """
        return f"https://api.yatta.top/hsr/assets/UI//equipment/medium/{self.light_cone_id}.png"



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from enka.gi import GenshinClient
from enka.hsr import HSRClient
from enka.zzz import ZZZClient

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator


@pytest.fixture
async def genshin_client() -> AsyncGenerator[GenshinClient]:
    async with GenshinClient() as client:
        yield client


@pytest.fixture
async def hsr_client() -> AsyncGenerator[HSRClient]:
    async with HSRClient() as client:
        yield client


@pytest.fixture
async def zzz_client() -> AsyncGenerator[ZZZClient]:
    async with ZZZClient() as client:
        yield client



from __future__ import annotations

import sys
from enum import IntEnum

from enka.constants.zzz import ELEMENT_ICON

if sys.version_info < (3, 11):
    from enum import Enum as StrEnum
else:
    from enum import StrEnum


class Language(StrEnum):
    """ZZZ language enum."""

    ENGLISH = "en"
    RUSSIAN = "ru"
    VIETNAMESE = "vi"
    THAI = "th"
    PORTUGUESE = "pt"
    KOREAN = "ko"
    JAPANESE = "ja"
    INDONESIAN = "id"
    FRENCH = "fr"
    ESPANOL = "es"
    GERMAN = "de"
    TRADITIONAL_CHINESE = "zh-tw"
    SIMPLIFIED_CHINESE = "zh-cn"


class SkillType(IntEnum):
    """ZZZ character skill type."""

    BASIC_ATK = 0
    SPECIAL_ATK = 1
    DASH = 2
    ULTIMATE = 3
    CORE_SKILL = 5
    ASSIST = 6


class StatType(IntEnum):
    """ZZZ stat type."""

    HP_BASE = 11101
    HP_PERCENT = 11102
    HP_FLAT = 11103

    ATK_BASE = 12101
    ATK_PERCENT = 12102
    ATK_FLAT = 12103

    IMPACT_BASE = 12201
    IMPACT_PERCENT = 12202

    DEF_BASE = 13101
    DEF_PERCENT = 13102
    DEF_FLAT = 13103

    CRIT_RATE_BASE = 20101
    CRIT_RATE_FLAT = 20103
    CRIT_DMG_BASE = 21101
    CRIT_DMG_FLAT = 21103

    PEN_RATIO_BASE = 23101
    PEN_RATIO_FLAT = 23103

    PEN_BASE = 23201
    PEN_FLAT = 23203

    ENERGY_REGEN_BASE = 30501
    ENERGY_REGEN_PERCENT = 30502
    ENERGY_REGEN_FLAT = 30503

    ANOMALY_PRO_BASE = 31201
    """Anomaly proficiency base."""
    ANOMALY_PRO_FLAT = 31203
    """Anomaly proficiency flat."""

    ANOMALY_MASTERY_BASE = 31401
    ANOMALY_MASTERY_PERCENT = 31402
    ANOMALY_MASTERY_FLAT = 31403

    PHYSICAL_DMG_BONUS_BASE = 31501
    PHYSICAL_DMG_BONUS_FLAT = 31503
    FIRE_DMG_BONUS_BASE = 31601
    FIRE_DMG_BONUS_FLAT = 31603
    ICE_DMG_BONUS_BASE = 31701
    ICE_DMG_BONUS_FLAT = 31703
    ELECTRIC_DMG_BONUS_BASE = 31801
    ELECTRIC_DMG_BONUS_FLAT = 31803
    ETHER_DMG_BONUS_BASE = 31901
    ETHER_DMG_BONUS_FLAT = 31903


class ProfessionType(StrEnum):
    """ZZZ agent speciality."""

    STUN = "Stun"
    ATTACK = "Attack"
    DEFENSE = "Defense"
    SUPPORT = "Support"
    ANOMALY = "Anomaly"
    UNKNOWN = "Unknown"


class Element(StrEnum):
    """ZZZ element type."""

    PHYSICAL = "Physics"
    FIRE = "Fire"
    ICE = "Ice"
    ELECTRIC = "Elec"
    ETHER = "Ether"
    FIRE_FROST = "FireFrost"
    UNKNOWN = "Unknown"

    @property
    def icon(self) -> str:
        """The element's icon."""
        return ELEMENT_ICON.format(element=self.value)



from __future__ import annotations

import math


def round_down(number: int | float, decimal_places: int) -> int | float:
    """Round down a number to a specified number of decimal places.

    Args:
        number: The number to round down.
        decimal_places: The number of decimal places to round down to.

    Returns:
        int | float: The rounded down number.
    """
    factor = 10**decimal_places
    return math.floor(number * factor) / factor



from __future__ import annotations

import sys
from enum import IntEnum

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from enum import Enum

    class StrEnum(str, Enum):
        pass


__all__ = ("Element", "Language", "Path", "RelicType", "StatType", "TraceType")


class Language(StrEnum):
    """Languages supported by the HSR Enka Network API."""

    ENGLISH = "en"
    RUSSIAN = "ru"
    VIETNAMESE = "vi"
    THAI = "th"
    PORTUGUESE = "pt"
    KOREAN = "ko"
    JAPANESE = "ja"
    INDOENSIAN = "id"
    FRENCH = "fr"
    ESPANOL = "es"
    GERMAN = "de"
    TRADITIONAL_CHINESE = "zh-tw"
    SIMPLIFIED_CHINESE = "zh-cn"


class RelicType(IntEnum):
    """HSR relic types."""

    HEAD = 1
    HAND = 2
    BODY = 3
    FOOT = 4
    ROPE = 5  # NECK
    ORBIT = 6  # OBJECT


class StatType(StrEnum):
    """HSR property types."""

    MAX_HP = "MaxHP"
    ATK = "Attack"
    DEF = "Defence"
    SPD = "Speed"

    CRIT_RATE = "CriticalChance"
    CRIT_DMG = "CriticalDamage"
    BREAK_EFFECT = "BreakDamageAddedRatio"
    BREAK_EFFECT_BASE = "BreakDamageAddedRatioBase"
    HEALING_BOOST = "HealRatio"

    MAX_ENERGY = "MaxSP"
    ENERGY_REGEN_RATE = "SPRatio"
    EFFECT_HIT_RATE = "StatusProbability"
    EFFECT_RES = "StatusResistance"
    CRIT_RATE_BASE = "CriticalChanceBase"
    CRIT_DMG_BASE = "CriticalDamageBase"
    HEALING_BOOST_BASE = "HealRatioBase"
    ENERGY_REGEN_RATE_BASE = "SPRatioBase"
    EFFECT_HIT_RATE_BASE = "StatusProbabilityBase"
    EFFECT_RES_BASE = "StatusResistanceBase"

    # Element dmg/resistance boosts
    PHYSICAL_DMG_BOOST = "PhysicalAddedRatio"
    PHYSICAL_RES_BOOST = "PhysicalResistance"
    FIRE_DMG_BOOST = "FireAddedRatio"
    FIRE_RES_BOOST = "FireResistance"
    ICE_DMG_BOOST = "IceAddedRatio"
    ICE_RES_BOOST = "IceResistance"
    LIGHTNING_DMG_BOOST = "ThunderAddedRatio"
    LIGHTNING_RES_BOOST = "ThunderResistance"
    WIND_DMG_BOOST = "WindAddedRatio"
    WIND_RES_BOOST = "WindResistance"
    QUANTUM_DMG_BOOST = "QuantumAddedRatio"
    QUANTUM_RES_BOOST = "QuantumResistance"
    IMAGINARY_DMG_BOOST = "ImaginaryAddedRatio"
    IMAGINARY_RES_BOOST = "ImaginaryResistance"

    BASE_HP = "BaseHP"
    HP_DELTA = "HPDelta"
    HP_BOOST = "HPAddedRatio"
    BASE_ATK = "BaseAttack"
    ATK_DELTA = "AttackDelta"
    ATK_BOOST = "AttackAddedRatio"
    BASE_DEF = "BaseDefence"
    DEF_DELTA = "DefenceDelta"
    DEF_BOOST = "DefenceAddedRatio"
    BASE_SPEED = "BaseSpeed"
    INCOMING_HEALING_BOOST = "HealTakenRatio"

    PHYSICAL_RES_DELTA = "PhysicalResistanceDelta"
    FIRE_RES_DELTA = "FireResistanceDelta"
    ICE_RES_DELTA = "IceResistanceDelta"
    THUNDER_RES_DELTA = "ThunderResistanceDelta"
    WIND_RES_DELTA = "WindResistanceDelta"
    QUANTUM_RES_DELTA = "QuantumResistanceDelta"
    IMAGINARY_RES_DELTA = "ImaginaryResistanceDelta"

    SPEED_DELTA = "SpeedDelta"


class Element(StrEnum):
    """HSR elements."""

    ICE = "Ice"
    FIRE = "Fire"
    WIND = "Wind"
    LIGHTNING = "Thunder"
    PHYSICAL = "Physical"
    QUANTUM = "Quantum"
    IMAGINARY = "Imaginary"


class Path(StrEnum):
    """HSR paths."""

    PRESERVATION = "Knight"
    THE_HUNT = "Rogue"
    ERUDITION = "Mage"
    NIHILITY = "Warlock"
    DESTRUCTION = "Warrior"
    HARMONY = "Shaman"
    ABUNDANCE = "Priest"
    REMEMBRANCE = "Memory"
    NONE = "None"


class TraceType(IntEnum):
    """HSR trace types."""

    UNKNOWN = 0
    """Set before post-processing"""
    SKILL = 2
    """Basic ATK, Skill, Ultimate, Talent, Technique"""
    TALENT = 3
    """Special talents (there are 3 of these)"""
    STAT = 1
    """Stat boost traces"""
    MEMOSPRITE = 4
    """Memosprite traces"""



from __future__ import annotations

from ..data import AssetData
from ..manager import AssetManager
from .file_paths import (
    AVATARS_PATH,
    EQUIPMENT_LEVEL_PATH,
    EQUIPMENTS_PATH,
    NAMECARDS_PATH,
    PROPERTY_PATH,
    TEXT_MAP_PATH,
    TITLES_PATH,
    WEAPON_LEVEL_PATH,
    WEAPON_STAR_PATH,
    WEAPONS_PATH,
)


class ZZZAssetManager(AssetManager):
    """Zenless Zone Zero asset manager."""

    def __init__(self) -> None:
        self.text_map = AssetData(TEXT_MAP_PATH)
        self.avatars = AssetData(AVATARS_PATH)
        self.equipments = AssetData(EQUIPMENTS_PATH)
        self.equipment_level = AssetData(EQUIPMENT_LEVEL_PATH)
        self.property = AssetData(PROPERTY_PATH)
        self.weapon_star = AssetData(WEAPON_STAR_PATH)
        self.weapon_level = AssetData(WEAPON_LEVEL_PATH)
        self.weapons = AssetData(WEAPONS_PATH)
        self.titles = AssetData(TITLES_PATH)
        self.namecards = AssetData(NAMECARDS_PATH)

        self._assets = (
            self.text_map,
            self.avatars,
            self.equipments,
            self.equipment_level,
            self.property,
            self.weapon_star,
            self.weapon_level,
            self.weapons,
            self.titles,
            self.namecards,
        )


ZZZ_ASSETS = ZZZAssetManager()



from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel, Field, computed_field, field_validator, model_validator

from ...constants.gi import ASCENSION_TO_MAX_LEVEL, DMG_BONUS_FIGHT_PROPS, PERCENT_STAT_TYPES
from ...enums.gi import Element, EquipmentType, FightPropType, ItemType, StatType
from ...errors import InvalidItemTypeError
from .costume import Costume
from .icon import Icon, Namecard

__all__ = ("Artifact", "Character", "Constellation", "FightProp", "Stat", "Talent", "Weapon")


class Stat(BaseModel):
    """Represents a stat.

    Attributes:
        type: The stat's type (e.g. FIGHT_PROP_HP, FIGHT_PROP_ATTACK).
        value: The stat's value.
        name: The stat's name.
    """

    type: StatType
    value: float
    name: str = ""

    @computed_field
    @property
    def is_percentage(self) -> bool:
        """Whether this stat is a percentage stat."""
        return self.type.name in PERCENT_STAT_TYPES

    @computed_field
    @property
    def formatted_value(self) -> str:
        """ "The formatted value of the stat."""
        if self.is_percentage:
            return f"{round(self.value, 1)}%"
        return str(round(self.value))


class FightProp(BaseModel):
    """Represents a character's stat (property.)

    Attributes:
        type: The fight prop's type (e.g. FIGHT_PROP_HP, FIGHT_PROP_ATTACK).
        value: The fight prop's value.
        name: The fight prop's name.
    """

    type: FightPropType | int
    value: float
    name: str = ""

    @field_validator("type", mode="before")
    @classmethod
    def __convert_type(cls, v: int) -> FightPropType | int:
        try:
            return FightPropType(v)
        except ValueError:
            return v

    @computed_field
    @property
    def is_percentage(self) -> bool:
        """Whether this stat is a percentage stat."""
        return isinstance(self.type, FightPropType) and self.type.name in PERCENT_STAT_TYPES

    @computed_field
    @property
    def formatted_value(self) -> str:
        """The formatted value of the stat."""
        if self.is_percentage:
            return f"{round(self.value * 100, 1)}%"
        return f"{round(self.value):,}"


class Artifact(BaseModel):
    """Represents an artifact.

    Attributes:
        id: The artifact's ID.
        main_stat_id: The main stat's ID.
        sub_stat_ids: The sub stats' IDs.
        level: The artifact's level.
        equip_type: The artifact's type (e.g. FLOWER, GOBLET).
        icon: The artifact's icon.
        item_type: The artifact's type.
        name: The artifact's name.
        rarity: The artifact's rarity.
        main_stat: The artifact's main stat.
        sub_stats: The artifact's sub stats.
        set_name: The artifact's set name.
    """

    id: int = Field(alias="itemId")
    main_stat_id: int = Field(alias="mainPropId")
    sub_stat_ids: list[int] = Field(alias="appendPropIdList", default_factory=list)
    level: int
    equip_type: EquipmentType = Field(alias="equipType")
    icon: str
    item_type: ItemType = Field(alias="itemType")
    name: str = Field(alias="nameTextMapHash")
    rarity: int = Field(alias="rankLevel")
    main_stat: Stat = Field(alias="reliquaryMainstat")
    sub_stats: list[Stat] = Field(alias="reliquarySubstats", default_factory=list)
    set_name: str = Field(alias="setNameTextMapHash")

    @field_validator("level", mode="before")
    @classmethod
    def _convert_level(cls, v: int) -> int:
        return v - 1

    @field_validator("icon", mode="before")
    @classmethod
    def _convert_icon(cls, v: str) -> str:
        return f"https://enka.network/ui/{v}.png"

    @field_validator("main_stat", mode="before")
    @classmethod
    def _convert_main_stat(cls, v: dict[str, Any]) -> Stat:
        return Stat(type=StatType(v["mainPropId"]), value=v["statValue"], name="")

    @field_validator("sub_stats", mode="before")
    @classmethod
    def _convert_sub_stats(cls, v: list[dict[str, Any]]) -> list[Stat]:
        return [
            Stat(type=StatType(stat["appendPropId"]), value=stat["statValue"], name="")
            for stat in v
        ]

    @field_validator("name", "set_name", mode="before")
    @classmethod
    def _stringify_text_map_hash(cls, v: str | int) -> str:
        return str(v)


class Weapon(BaseModel):
    """Represents a weapon.

    Attributes:
        item_id: The weapon's ID.
        refinement: The weapon's refinement level (1~5).
        level: The weapon's level.
        ascension: The weapon's ascension level.
        icon: The weapon's icon.
        name: The weapon's name.
        rarity: The weapon's rarity.
        stats: The weapon's stats.
    """

    item_id: int = Field(alias="itemId")
    refinement: Literal[1, 2, 3, 4, 5] = Field(1, alias="affixMap")
    level: int
    ascension: Literal[0, 1, 2, 3, 4, 5, 6] = Field(0, alias="promoteLevel")
    icon: str
    item_type: ItemType = Field(alias="itemType")
    name: str = Field(alias="nameTextMapHash")
    rarity: int = Field(alias="rankLevel")
    stats: list[Stat] = Field(alias="weaponStats")

    @computed_field
    @property
    def max_level(self) -> int:
        return ASCENSION_TO_MAX_LEVEL[self.ascension]

    @field_validator("refinement", mode="before")
    @classmethod
    def _extract_refinement(cls, v: dict[str, int]) -> int:
        return next(iter(v.values())) + 1

    @field_validator("icon", mode="before")
    @classmethod
    def _convert_icon(cls, v: str) -> str:
        return f"https://enka.network/ui/{v}.png"

    @field_validator("stats", mode="before")
    @classmethod
    def _convert_stats(cls, v: list[dict[str, Any]]) -> list[Stat]:
        return [
            Stat(type=StatType(stat["appendPropId"]), value=stat["statValue"], name="")
            for stat in v
            if stat["statValue"] != 0
        ]

    @field_validator("name", mode="before")
    @classmethod
    def _stringify_text_map_hash(cls, v: str | int) -> str:
        return str(v)


class Constellation(BaseModel):
    """Represents a character's constellation.

    Attributes:
        id: The constellation's ID.
        name: The constellation's name.
        icon: The constellation's icon.
        unlocked: Whether the constellation is unlocked.
    """

    id: int
    name: str = ""
    icon: str = ""
    unlocked: bool


class Talent(BaseModel):
    """Represents a character's talent.

    Attributes:
        id: The talent's ID.
        level: The talent's level.
        name: The talent's name.
        icon: The talent's icon.
        is_upgraded: Whether the talent's level is being upgraded by a constellation.
    """

    id: int
    level: int
    name: str = ""
    icon: str = ""
    is_upgraded: bool = False


class Character(BaseModel):
    """Represents a character.

    Attributes:
        id: The character's ID.
        artifacts: The character's artifacts.
        weapon: The character's weapon.
        stats: The character's stats.
        constellations: The character's unlocked constellations.
        talents: The character's talents.
        ascension: The character's ascension level.
        level: The character's level.
        skill_depot_id: The character's skill depot ID.
        name: The character's name.
        talent_extra_level_map: The map of character's extra talent levels, this is only used internally, the wrapper will handle this.
        icon: The character's icon.
        friendship_level: The character's friendship level (1~10).
        element: The character's element.
        talent_order: The character's talent order.
            1. Normal attack
            2. Elemental skill
            3. Elemental burst
        rarity: The character's rarity (4~5).
        max_level: The character's max level.
        highest_dmg_bonus_stat: The character's highest damage bonus stat.
        namecard: The character's namecard. Travelers don't have namecards.
        costume: The character's costume, if any.
        costume_id: The character's costume's ID, if any.
        constellations_unlocked: The number of constellations unlocked.
    """

    id: int = Field(alias="avatarId")
    artifacts: list[Artifact]
    weapon: Weapon
    stats: dict[FightPropType | int, FightProp] = Field(alias="fightPropMap")
    constellations: list[Constellation] = Field(alias="talentIdList", default_factory=list)
    talents: list[Talent] = Field(alias="skillLevelMap")
    ascension: Literal[0, 1, 2, 3, 4, 5, 6]
    level: int
    skill_depot_id: int = Field(alias="skillDepotId")
    talent_extra_level_map: dict[str, int] | None = Field(None, alias="proudSkillExtraLevelMap")
    friendship_level: int = Field(alias="friendshipLevel")

    name: str = ""
    icon: Icon = Icon(side_icon_ui_path="")
    element: Element = Element.ANEMO
    talent_order: list[int] = Field(default_factory=list)
    rarity: int = 0
    namecard: Namecard | None = None
    costume: Costume | None = None
    costume_id: int | None = Field(None, alias="costumeId")

    model_config = {"arbitrary_types_allowed": True}

    @computed_field
    @property
    def max_level(self) -> Literal[20, 40, 50, 60, 70, 80, 90]:
        """The character's max level."""
        return ASCENSION_TO_MAX_LEVEL[self.ascension]

    @computed_field
    @property
    def highest_dmg_bonus_stat(self) -> FightProp:
        """The character's highest damage bonus stat.

        Returns the highest stat value from the damage bonus stats (elemental damage bonus, physical damage bonus, etc.).
        """
        return max(
            (
                stat
                for stat in self.stats.values()
                if isinstance(stat.type, FightPropType) and stat.type.name in DMG_BONUS_FIGHT_PROPS
            ),
            key=lambda stat: stat.value,
        )

    @computed_field
    @property
    def specialized_stat(self) -> FightProp:
        """The character's specialized stat

        Returns the highest stat value from the specialized stats (elemental damage bonus and healing bonus).
        """
        specialized_stats = [*list(DMG_BONUS_FIGHT_PROPS), FightPropType.FIGHT_PROP_HEAL_ADD.name]
        return max(
            (
                stat
                for stat in self.stats.values()
                if isinstance(stat.type, FightPropType) and stat.type.name in specialized_stats
            ),
            key=lambda stat: stat.value,
        )

    @computed_field
    @property
    def constellations_unlocked(self) -> int:
        """The number of constellations unlocked."""
        return len([c for c in self.constellations if c.unlocked])

    @field_validator("ascension", mode="before")
    @classmethod
    def _intify_ascension(cls, v: str) -> int:
        return int(v)

    @field_validator("stats", mode="before")
    @classmethod
    def _convert_stats(cls, v: dict[str, float]) -> dict[FightPropType | int, FightProp]:
        result: dict[FightPropType | int, FightProp] = {}
        for k, value in v.items():
            if k.isdigit() and int(k) in FightPropType._value2member_map_:
                result[FightPropType(int(k))] = FightProp(
                    type=FightPropType(int(k)), value=value, name=""
                )
            else:
                result[int(k)] = FightProp(type=int(k), value=value, name="")
        return result

    @field_validator("constellations", mode="before")
    @classmethod
    def _convert_constellations(cls, v: list[int]) -> list[Constellation]:
        return [
            Constellation(id=constellation_id, name="", icon="", unlocked=True)
            for constellation_id in v
        ]

    @field_validator("talents", mode="before")
    @classmethod
    def _convert_talents(cls, v: dict[str, int]) -> list[Talent]:
        return [Talent(id=int(k), level=v, name="", icon="") for k, v in v.items()]

    @field_validator("weapon", mode="before")
    @classmethod
    def _flatten_weapon_data(cls, v: dict[str, Any]) -> dict[str, Any]:
        v.update(v["weapon"])
        v.update(v["flat"])
        v.pop("weapon")
        v.pop("flat")
        return v

    @field_validator("artifacts", mode="before")
    @classmethod
    def _flatten_artifacts_data(cls, v: list[dict[str, Any]]) -> list[dict[str, Any]]:
        for artifact in v:
            artifact.update(artifact["reliquary"])
            artifact.update(artifact["flat"])
            artifact.pop("reliquary")
            artifact.pop("flat")

        return v

    @model_validator(mode="before")
    @classmethod
    def _transform_values(cls, v: dict[str, Any]) -> dict[str, Any]:
        # convert prop map to level and ascension
        prop_map = v["propMap"]
        try:
            v["level"] = prop_map["4001"]["val"]
        except KeyError:
            v["level"] = 1
        try:
            v["ascension"] = prop_map["1002"]["val"]
        except KeyError:
            v["ascension"] = 0

        # convert equipment list to weapon and artifacts
        equip_list = v["equipList"]
        v["artifacts"] = []

        for equipment in equip_list:
            if "weapon" in equipment:
                v["weapon"] = equipment
            elif "reliquary" in equipment:
                v["artifacts"].append(equipment)
            else:
                raise InvalidItemTypeError

        # friendship level
        v["friendshipLevel"] = v["fetterInfo"]["expLevel"]

        return v



from __future__ import annotations

from pathlib import Path
from typing import Final

ASSET_PATH = Path(".enka_py/assets/hsr")

TEXT_MAP_PATH = ASSET_PATH / "text_map.json"
CHARACTER_DATA_PATH = ASSET_PATH / "characters.json"
LIGHT_CONE_DATA_PATH = ASSET_PATH / "light_cone.json"
RELIC_DATA_PATH = ASSET_PATH / "relic.json"
SKILL_TREE_DATA_PATH = ASSET_PATH / "skill_tree.json"
META_DATA_PATH = ASSET_PATH / "promotions.json"
AVATAR_DATA_PATH = ASSET_PATH / "avatars.json"
PROPERTY_CONFIG_PATH = ASSET_PATH / "property_config.json"
EIDOLON_DATA_PATH = ASSET_PATH / "eidolons.json"

ENKA_API_DOCS = "https://raw.githubusercontent.com/pizza-studio/EnkaDBGenerator/main/Sources/EnkaDBFiles/Resources/Specimen/HSR"
ENKA_PY_ASSETS = "https://raw.githubusercontent.com/seriaati/enka-py-assets/main/data/hsr"

PATH_TO_SOURCE: Final[dict[Path, str]] = {
    CHARACTER_DATA_PATH: f"{ENKA_API_DOCS}/honker_characters.json",
    LIGHT_CONE_DATA_PATH: f"{ENKA_API_DOCS}/honker_weps.json",
    RELIC_DATA_PATH: f"{ENKA_API_DOCS}/honker_relics.json",
    META_DATA_PATH: f"{ENKA_API_DOCS}/honker_meta.json",
    AVATAR_DATA_PATH: f"{ENKA_API_DOCS}/honker_avatars.json",
    EIDOLON_DATA_PATH: f"{ENKA_API_DOCS}/honker_ranks.json",
    SKILL_TREE_DATA_PATH: f"{ENKA_PY_ASSETS}/skill_tree.json",
    PROPERTY_CONFIG_PATH: f"{ENKA_PY_ASSETS}/property_config.json",
    TEXT_MAP_PATH: f"{ENKA_PY_ASSETS}/hsr.json",
}



from __future__ import annotations

from pathlib import Path
from typing import Final

ASSET_PATH = Path(".enka_py/assets/zzz")

AVATARS_PATH = ASSET_PATH / "avatars.json"
TEXT_MAP_PATH = ASSET_PATH / "text_map.json"
EQUIPMENT_LEVEL_PATH = ASSET_PATH / "equipment_level.json"
EQUIPMENTS_PATH = ASSET_PATH / "equipments.json"
PROPERTY_PATH = ASSET_PATH / "property.json"
WEAPON_LEVEL_PATH = ASSET_PATH / "weapon_level.json"
WEAPON_STAR_PATH = ASSET_PATH / "weapon_star.json"
WEAPONS_PATH = ASSET_PATH / "weapons.json"
TITLES_PATH = ASSET_PATH / "titles.json"
NAMECARDS_PATH = ASSET_PATH / "namecards.json"

ENKA_API_DOCS = "https://raw.githubusercontent.com/EnkaNetwork/API-docs/master/store/zzz"
ENKA_PY_ASSETS = "https://raw.githubusercontent.com/seriaati/enka-py-assets/main/data/zzz"

PATH_TO_SOURCE: Final[dict[Path, str]] = {
    AVATARS_PATH: f"{ENKA_API_DOCS}/avatars.json",
    TEXT_MAP_PATH: f"{ENKA_API_DOCS}/locs.json",
    EQUIPMENTS_PATH: f"{ENKA_API_DOCS}/equipments.json",
    PROPERTY_PATH: f"{ENKA_API_DOCS}/property.json",
    WEAPONS_PATH: f"{ENKA_API_DOCS}/weapons.json",
    EQUIPMENT_LEVEL_PATH: f"{ENKA_PY_ASSETS}/equipment_level.json",
    WEAPON_LEVEL_PATH: f"{ENKA_PY_ASSETS}/weapon_level.json",
    WEAPON_STAR_PATH: f"{ENKA_PY_ASSETS}/weapon_star.json",
    TITLES_PATH: f"{ENKA_PY_ASSETS}/titles.json",
    NAMECARDS_PATH: f"{ENKA_PY_ASSETS}/namecards.json",
}



from __future__ import annotations

from typing import Literal

RARITY_MAP: dict[int, Literal["S", "A", "B"]] = {4: "S", 3: "A", 2: "B"}

ICON_FOLDER = "https://raw.githubusercontent.com/seriaati/enka-py-assets/refs/heads/main/icons/zzz"
ELEMENT_ICON = f"{ICON_FOLDER}/element/{{element}}.png"



from __future__ import annotations

from typing import Literal

__all__ = (
    "ASCENSION_TO_MAX_LEVEL",
    "CHARACTER_RARITY_MAP",
    "DMG_BONUS_FIGHT_PROPS",
    "PERCENT_STAT_TYPES",
)

PERCENT_STAT_TYPES = {
    "FIGHT_PROP_HP_PERCENT",
    "FIGHT_PROP_ATTACK_PERCENT",
    "FIGHT_PROP_DEFENSE_PERCENT",
    "FIGHT_PROP_CRITICAL",
    "FIGHT_PROP_CRITICAL_HURT",
    "FIGHT_PROP_CHARGE_EFFICIENCY",
    "FIGHT_PROP_HEAL_ADD",
    "FIGHT_PROP_PHYSICAL_ADD_HURT",
    "FIGHT_PROP_FIRE_ADD_HURT",
    "FIGHT_PROP_ELEC_ADD_HURT",
    "FIGHT_PROP_WATER_ADD_HURT",
    "FIGHT_PROP_WIND_ADD_HURT",
    "FIGHT_PROP_ROCK_ADD_HURT",
    "FIGHT_PROP_ICE_ADD_HURT",
    "FIGHT_PROP_GRASS_ADD_HURT",
}

CHARACTER_RARITY_MAP: dict[str, Literal[4, 5]] = {
    "QUALITY_ORANGE": 5,
    "QUALITY_ORANGE_SP": 5,
    "QUALITY_PURPLE": 4,
}

ASCENSION_TO_MAX_LEVEL: dict[Literal[0, 1, 2, 3, 4, 5, 6], Literal[20, 40, 50, 60, 70, 80, 90]] = {
    0: 20,
    1: 40,
    2: 50,
    3: 60,
    4: 70,
    5: 80,
    6: 90,
}

DMG_BONUS_FIGHT_PROPS = {
    "FIGHT_PROP_PHYSICAL_ADD_HURT",
    "FIGHT_PROP_FIRE_ADD_HURT",
    "FIGHT_PROP_ELEC_ADD_HURT",
    "FIGHT_PROP_WATER_ADD_HURT",
    "FIGHT_PROP_WIND_ADD_HURT",
    "FIGHT_PROP_ROCK_ADD_HURT",
    "FIGHT_PROP_ICE_ADD_HURT",
    "FIGHT_PROP_GRASS_ADD_HURT",
}



from __future__ import annotations

from pathlib import Path

ASSET_PATH = Path(".enka_py/assets")

TEXT_MAP_PATH = ASSET_PATH / "text_map.json"
CHARACTER_DATA_PATH = ASSET_PATH / "characters.json"
NAMECARD_DATA_PATH = ASSET_PATH / "namecards.json"
CONSTS_DATA_PATH = ASSET_PATH / "consts.json"
TALENTS_DATA_PATH = ASSET_PATH / "talents.json"
PFPS_DATA_PATH = ASSET_PATH / "pfps.json"

ENKA_API_DOCS = "https://raw.githubusercontent.com/pizza-studio/EnkaDBGenerator/main/Sources/EnkaDBFiles/Resources/Specimen/GI"
ENKA_PY_ASSETS = "https://raw.githubusercontent.com/seriaati/enka-py-assets/main/data"

PATH_TO_SOURCE = {
    TEXT_MAP_PATH: f"{ENKA_PY_ASSETS}/text_map.json",
    CHARACTER_DATA_PATH: f"{ENKA_PY_ASSETS}/characters.json",
    CONSTS_DATA_PATH: f"{ENKA_PY_ASSETS}/consts.json",
    TALENTS_DATA_PATH: f"{ENKA_PY_ASSETS}/talents.json",
    PFPS_DATA_PATH: f"{ENKA_API_DOCS}/pfps.json",
    NAMECARD_DATA_PATH: f"{ENKA_API_DOCS}/namecards.json",
}



from __future__ import annotations

from pydantic import BaseModel

__all__ = ("Owner", "OwnerProfile")


class OwnerProfile(BaseModel):
    """Represents an Enka account owner's profile information.

    Attributes:
        bio (str): The owner's bio.
        avatar (str, optional): The owner's avatar.
    """

    bio: str
    avatar: str | None = None


class Owner(BaseModel):
    """Represents an Enka account owner.

    Attributes:
        hash (str): The owner's hash.
        username (str): The owner's username.
        profile (:class:`OwnerProfile`): The owner's profile information.
        id (int): The owner's ID.
    """

    hash: str
    username: str
    profile: OwnerProfile
    id: int



from __future__ import annotations

from typing import Any

from deprecated import deprecated
from pydantic import BaseModel, Field, model_validator

__all__ = ("Medal", "Namecard", "Player", "Title")


class Medal(BaseModel):
    """Represents a medal obtained by a player in ZZZ.

    Attributes:
        value: The value associated with the medal.
        type: The type identifier of the medal.
        icon_id: The identifier for the medal's icon.
    """

    value: int = Field(alias="Value")
    type: int = Field(alias="MedalType")
    icon_id: int = Field(alias="MedalIcon")


class Title(BaseModel):
    """Represents a title obtained by a player in ZZZ.

    Attributes:
        id: The unique identifier for the title.
        name: The name of the title (e.g., "Cooked More Than a Hamburger").
        color1: The primary color associated with the title (e.g., "#e6e9ea").
        color2: The secondary color associated with the title (e.g., "#8ea3ae").
    """

    id: int = Field(alias="Title")

    # Fields that are not in the API response
    text: str = ""
    color1: str = ""
    color2: str = ""


class Namecard(BaseModel):
    """Represents a namecard obtained by a player in ZZZ.

    Attributes:
        id: The unique identifier for the name card.
        icon: The icon associated with the name card.
    """

    id: int

    # Fields that are not in the API response
    icon: str = ""


class Player(BaseModel):
    """Represents a ZZZ player's profile information.

    Attributes:
        medals: List of medals obtained by the player.
        nickname: The player's in-game nickname.
        avatar_id: The ID of the player's selected avatar.
        uid: The player's unique identifier.
        level: The player's current level.
        signature: The player's custom signature or description.
        title: The player's title.
        id: The player's profile ID.
        namecard_id: The ID of the player's namecard.
        namecard: The player's namecard.
    """

    medals: list[Medal] = Field(alias="MedalList")

    nickname: str = Field(alias="Nickname")
    avatar_id: int = Field(alias="AvatarId")
    uid: int = Field(alias="Uid")
    level: int = Field(alias="Level")
    signature: str = Field(alias="Desc")

    title: Title = Field(alias="TitleInfo")
    id: int = Field(alias="ProfileId")
    namecard_id: int = Field(alias="CallingCardId")

    # Fields that are not in the API response
    namecard: Namecard = Field(default=None)  # pyright: ignore[reportAssignmentType]

    @model_validator(mode="before")
    @classmethod
    def __unnest_info(cls, value: dict[str, Any]) -> dict[str, Any]:
        info = value.pop("ProfileDetail", {})
        value.update(info)
        return value

    @property
    @deprecated(reason="Use `title.id` instead.")
    def title_id(self) -> int:
        """(Deprecated) The ID of the player's selected title."""
        return self.title.id



from __future__ import annotations

import datetime
from typing import Any, Literal

from pydantic import BaseModel, Field, computed_field, field_validator, model_validator

from ...constants.zzz import RARITY_MAP
from ...enums.zzz import Element, ProfessionType, SkillType, StatType
from ...models.zzz.icon import AgentIcon

__all__ = ("Agent", "AgentColor", "AgentSkill", "DriveDisc", "DriveDiscStat", "Stat", "WEngine")


def _to_formatted_value(value: int, type_: StatType, format_: str) -> str:
    if type_ is StatType.ENERGY_REGEN_BASE:
        return f"{value / 100}"

    if "%" in format_:
        return f"{round(value / 100, 1)}%"
    return str(int(value))


class Stat(BaseModel):
    """Represents a agent's or W-Engine's stat.

    Attributes:
        type: The type of the stat.
        value: The value of the stat.
        name: The name of the stat.
        format: The format specifier.
    """

    type: StatType
    value: int
    name: str
    format: str

    @computed_field
    @property
    def formatted_value(self) -> str:
        """The formatted value of the stat."""
        return _to_formatted_value(self.value, self.type, self.format)


class DriveDiscStat(Stat):
    """Represents a drive disc's stat.

    Attributes:
        type: The type of the stat.
        roll_times: The number of times the stat has been rolled.
        value: The value of the stat.
        name: The name of the stat.
        format: The format specifier.
    """

    type: StatType = Field(alias="PropertyId")
    roll_times: int = Field(alias="PropertyLevel")
    value: int = Field(alias="PropertyValue")

    # Fields that are not in the API response
    name: str = ""
    format: str = ""

    @computed_field
    @property
    def formatted_value(self) -> str:
        """The formatted value of the stat."""
        return _to_formatted_value(self.value, self.type, self.format)


class DriveDisc(BaseModel):
    """Represents a drive disc.

    Attributes:
        slot: The slot number of the drive disc.
        id: The ID of the drive disc.
        uid: The unique identifier of the drive disc.
        level: The level of the drive disc.
        roll_times: The number of times random stats are being rolled.
        main_stat: The main stat of the drive disc.
        sub_stats: List of sub-stats of the drive disc.
        is_locked: Whether the drive disc is marked as locked in-game.
        is_trash: Whether the drive disc is marked as trash in-game.
        rarity_num: The rarity number of the drive disc.
        set_id: The set ID of the drive disc.
    """

    slot: Literal[1, 2, 3, 4, 5, 6] = Field(alias="Slot")

    id: int = Field(alias="Id")
    uid: int = Field(alias="Uid")
    level: int = Field(alias="Level")
    roll_times: int = Field(alias="BreakLevel")

    main_stat: DriveDiscStat = Field(alias="MainPropertyList")
    sub_stats: list[DriveDiscStat] = Field(alias="RandomPropertyList")

    is_locked: bool = Field(alias="IsLocked")
    is_trash: bool = Field(alias="IsTrash")

    # Fields that are not in the API response
    rarity_num: int = Field(default=0)
    set_id: int = Field(default=0)

    @field_validator("main_stat", mode="before")
    @classmethod
    def __get_first_main_stat_value(cls, value: list[dict[str, Any]]) -> dict[str, Any]:
        return value[0]

    @model_validator(mode="before")
    @classmethod
    def __unnest_info(cls, value: dict[str, Any]) -> dict[str, Any]:
        info = value.pop("Equipment")
        value.update(info)
        return value

    @computed_field
    @property
    def rarity(self) -> Literal["S", "A", "B", "?"]:
        """The drive disc's rarity."""
        return RARITY_MAP.get(self.rarity_num, "?")


class WEngine(BaseModel):
    """Represents a W-Engine.

    Attributes:
        id: The ID of the W-Engine.
        uid: The unique identifier of the W-Engine.
        level: The level of the W-Engine.
        modification: The modification level of the W-Engine.
        phase: The phase level of the W-Engine.
        is_locked: Whether the W-Engine is marked as locked in-game.
        rarity_num: The rarity number of the W-Engine.
        name: The name of the W-Engine.
        specialty: The specialty of the W-Engine.
        icon: The icon of the W-Engine.
        main_stat: The main stat of the W-Engine.
        sub_stat: The sub-stat of the W-Engine.
    """

    id: int = Field(alias="Id")
    uid: int = Field(alias="Uid")
    level: int = Field(alias="Level")
    modification: Literal[0, 1, 2, 3, 4, 5] = Field(alias="BreakLevel")
    phase: Literal[1, 2, 3, 4, 5] = Field(alias="UpgradeLevel")

    is_locked: bool = Field(alias="IsLocked")

    # Fields that are not in the API response
    rarity_num: int = Field(default=0)
    name: str = ""
    specialty: ProfessionType = ProfessionType.UNKNOWN
    icon: str = ""
    main_stat: Stat = Field(default=None)  # pyright: ignore[reportAssignmentType]
    sub_stat: Stat = Field(default=None)  # pyright: ignore[reportAssignmentType]

    @computed_field
    @property
    def rarity(self) -> Literal["S", "A", "B", "?"]:
        """The W-Engine's rarity."""
        return RARITY_MAP.get(self.rarity_num, "?")


class AgentSkill(BaseModel):
    """Represents a ZZZ agent's skill.

    Attributes:
        level: The level of the skill.
        type: The type of the skill.
    """

    level: int = Field(alias="Level")
    type: SkillType = Field(alias="Index")


class AgentColor(BaseModel):
    """Represents a ZZZ agent's color scheme.

    Attributes:
        accent: The accent color of the agent.
        mindscape: The mindscape color of the agent.
    """

    accent: str = Field(alias="Accent")
    mindscape: str = Field(alias="Mindscape")


class Agent(BaseModel):
    """Represents a ZZZ agent (character).

    Attributes:
        id: The ID of the agent.
        uid: The unique identifier of the agent.
        level: The level of the agent.
        promotion: The promotion level of the agent.
        mindscape: The mindscape level of the agent.
        skin_id: The ID of the agent's skin.
        core_skill_level_num: The core skill level number of the agent.
        is_sig_engine_effect_on: Whether the signature engine effect is on.
        obtained_at: The time when the agent was obtained, in server timezone.
        skills: List of skills for the agent.
        discs: List of drive discs equipped by the agent.
        w_engine: The W-Engine associated with the agent.
        name: The name of the agent.
        rarity_num: The rarity number of the agent.
        elements: List of elements associated with the agent.
        icon: The icon of the agent.
        sig_engine_id: The ID of the signature W-Engine.
        color: The color scheme of the agent.
        highlight_stats: List of stats that are highlighted in the agent menu.
        stats: Dictionary of stats for the agent.
    """

    id: int = Field(alias="Id")
    level: int = Field(alias="Level")
    promotion: Literal[1, 2, 3, 4, 5, 6] = Field(alias="PromotionLevel")
    mindscape: Literal[0, 1, 2, 3, 4, 5, 6] = Field(alias="TalentLevel")
    skin_id: int = Field(alias="SkinId")
    core_skill_level_num: int = Field(alias="CoreSkillEnhancement")

    is_sig_engine_effect_on: bool | None = Field(alias="WeaponEffectState")
    obtained_at: datetime.datetime = Field(alias="ObtainmentTimestamp")

    skills: list[AgentSkill] = Field(alias="SkillLevelList")
    discs: list[DriveDisc] = Field(alias="EquippedList")
    w_engine: WEngine | None = Field(None, alias="Weapon")

    # Fields that are not in the API response
    name: str = Field(default="")
    rarity_num: int = Field(default=0)
    elements: list[Element] = Field(default_factory=list)
    icon: AgentIcon = Field(default=None)  # pyright: ignore[reportAssignmentType]
    sig_engine_id: int = Field(default=0)
    color: AgentColor = Field(default=None)  # pyright: ignore[reportAssignmentType]
    highlight_stats: list[StatType] = Field(default_factory=list)
    stats: dict[StatType, Stat] = Field(default_factory=dict)

    @field_validator("is_sig_engine_effect_on", mode="before")
    @classmethod
    def __convert_is_sig_engine_effect_on(cls, value: int) -> bool | None:
        # [0: None, 1: OFF, 2: ON]
        return {0: None, 1: False, 2: True}[value]

    @computed_field
    @property
    def rarity(self) -> Literal["S", "A", "B", "?"]:
        """The agent's rarity."""
        return RARITY_MAP.get(self.rarity_num, "?")

    @computed_field
    @property
    def core_skill_level(self) -> Literal["A", "B", "C", "D", "E", "F"]:
        """The agent's core skill level."""
        return "ABCDEF"[self.core_skill_level_num - 1]  # pyright: ignore[reportReturnType]



from __future__ import annotations

from pydantic import BaseModel, computed_field

__all__ = ("AgentIcon",)

ENKA_BASE = "https://enka.network/ui/zzz/{filename}.png"
HAKUSHIN_BASE = "https://api.hakush.in/zzz/UI/{filename}.webp"


class AgentIcon(BaseModel):
    """Represents a ZZZ character's icon.

    Attributes:
        filename: Should be "IconRoleXX", parsed from "image" property.
    """

    filename: str

    @computed_field
    @property
    def round(self) -> str:
        """Character's icon in circle/rounded shape.

        e.g. https://enka.network/ui/zzz/IconRoleCircle13.png
        """
        return ENKA_BASE.format(filename=self.filename.replace("IconRole", "IconRoleCircle"))

    @computed_field
    @property
    def image(self) -> str:
        """Character's full image.

        e.g. https://enka.network/ui/zzz/IconRole13.png
        """
        return ENKA_BASE.format(filename=self.filename)

    @computed_field
    @property
    def select(self) -> str:
        """Character's icon in the select screen.

        e.g. https://api.hakush.in/zzz/UI/IconRoleSelect13.webp
        """
        return HAKUSHIN_BASE.format(filename=self.filename.replace("IconRole", "IconRoleSelect"))



from __future__ import annotations

import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from enum import Enum

    class StrEnum(str, Enum):
        pass


__all__ = ("Game",)


class Game(StrEnum):
    """Game."""

    GI = "gi"
    HSR = "hsr"
    ZZZ = "zzz"



from __future__ import annotations

import pytest

from enka.hsr import HSRClient, Language


async def test_update_assets(hsr_client: HSRClient) -> None:
    await hsr_client.update_assets()


async def test_fetch_showcase(hsr_client: HSRClient) -> None:
    await hsr_client.fetch_showcase("809162009")


async def test_trailblazer_showcase(hsr_client: HSRClient) -> None:
    await hsr_client.fetch_showcase("123456786")


@pytest.mark.parametrize("lang", list(Language))
async def test_langs(lang: Language) -> None:
    async with HSRClient(lang) as api:
        await api.fetch_showcase("809162009")


async def test_low_level_acc(hsr_client: HSRClient) -> None:
    await hsr_client.fetch_showcase("829702635")


async def test_owner_and_builds(hsr_client: HSRClient) -> None:
    showcase = await hsr_client.fetch_showcase("809162009")
    assert showcase.owner is not None
    await hsr_client.fetch_builds(showcase.owner)


async def test_raw_and_parse(hsr_client: HSRClient) -> None:
    raw = await hsr_client.fetch_showcase("809162009", raw=True)
    hsr_client.parse_showcase(raw)


async def test_empty_substat_list(hsr_client: HSRClient) -> None:
    await hsr_client.fetch_showcase("800724088")



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any, Final, Literal, overload

from loguru import logger

from ..assets.data import TextMap
from ..assets.gi.manager import GI_ASSETS
from ..constants.gi import CHARACTER_RARITY_MAP
from ..enums.enum import Game
from ..enums.gi import Element, FightPropType, Language
from ..models.gi import Constellation, Costume, Icon, Namecard, ShowcaseResponse
from ..models.gi.build import Build
from .base import BaseClient

if TYPE_CHECKING:
    from ..models.enka.owner import Owner
    from ..models.gi.character import Artifact, Character, Weapon
    from ..models.gi.player import Player, ShowcaseCharacter
    from .cache import BaseTTLCache

__all__ = ("GenshinClient",)

API_URL: Final[str] = "https://enka.network/api/uid/{uid}"


class GenshinClient(BaseClient):
    """The main client to interact with the Enka Network Genshin Impact API.

    Args:
        lang: The language to use for the client, defaults to Language.ENGLISH.
        headers: The headers to use for the client, defaults to None.
        cache: The cache to use for the client, defaults to None.
    """

    def __init__(
        self,
        lang: Language | str = Language.ENGLISH,
        *,
        headers: dict[str, Any] | None = None,
        cache: BaseTTLCache | None = None,
    ) -> None:
        super().__init__(Game.GI, headers=headers, cache=cache)

        self._lang = self._convert_lang(lang)
        self._assets = GI_ASSETS

    def _convert_lang(self, lang: Language | str) -> Language:
        if not isinstance(lang, Language):
            try:
                lang = Language(lang)
            except ValueError as e:
                available_langs = ", ".join(lang.value for lang in Language)
                msg = f"Invalid language: {lang}, must be one of {available_langs}"
                raise ValueError(msg) from e

        return lang

    @property
    def lang(self) -> Language:
        return self._lang

    @lang.setter
    def lang(self, lang: Language | str) -> None:
        self._lang = self._convert_lang(lang)
        self._text_map = TextMap(self._lang, self._assets.text_map)

    async def __aenter__(self) -> GenshinClient:
        await self.start()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return await super().__aexit__(exc_type, exc_val, exc_tb)

    def _post_process_player(self, player: Player) -> None:
        # namecard
        namecard_icon = self._assets.namecard_data[str(player.namecard_id)]["icon"]
        player.namecard = Namecard(ui_path=namecard_icon)

        # profile picture
        profile_picture_id = str(player.profile_picture_id)
        if len(profile_picture_id) == 8:
            profile_picture_icon = self._assets.character_data[profile_picture_id]["SideIconName"]
        else:
            # pfps data always return circle icon, so we need to process it to side icon for the `Icon` object.
            profile_picture_icon = (
                self._assets.pfps_data[profile_picture_id]["iconPath"]
                .replace("AvatarIcon", "AvatarIcon_Side")
                .replace("_Circle", "")
            )
        player.profile_picture_icon = Icon(
            side_icon_ui_path=profile_picture_icon, is_costume="Costume" in profile_picture_icon
        )

    def _post_process_showcase_character(self, showcase_character: ShowcaseCharacter) -> None:
        if showcase_character.costume_id is None:
            return

        # costume
        costume_data = self._assets.character_data[str(showcase_character.id)]["Costumes"]
        if costume_data is not None:
            showcase_character.costume = Costume(
                id=showcase_character.costume_id,
                data=costume_data[str(showcase_character.costume_id)],
            )

    def _post_process_weapon(self, weapon: Weapon) -> None:
        weapon.name = self._text_map[weapon.name]
        for stat in weapon.stats:
            stat.name = self._text_map[stat.type.value]

    def _post_process_talents(self, c: Character, data: dict[str, Any]) -> None:
        for talent in c.talents:
            talent_data = self._assets.talents_data[str(talent.id)]
            talent.name = self._text_map[talent_data["nameTextMapHash"]]
            talent.icon = f"https://enka.network/ui/{talent_data['icon']}.png"
            if c.talent_extra_level_map:
                proud_map: dict[str, int] = data["ProudMap"]
                proud_id = proud_map.get(str(talent.id))
                if proud_id is None:
                    continue
                if extra_level := c.talent_extra_level_map.get(str(proud_id), 0):
                    talent.level += extra_level
                    talent.is_upgraded = True

    def _post_process_artifact(self, artifact: Artifact) -> None:
        artifact.name = self._text_map[artifact.name]
        artifact.set_name = self._text_map[artifact.set_name]
        artifact.main_stat.name = self._text_map[artifact.main_stat.type.value]
        for stat in artifact.sub_stats:
            stat.name = self._text_map[stat.type.value]

    def _post_process_constellations(self, c: Character, data: dict[str, Any]) -> None:
        all_consts: list[dict[str, str]] = []
        for const_id, const in self._assets.consts_data.items():
            if const["icon"] not in data["Consts"]:
                continue
            const["id"] = const_id
            all_consts.append(const)

        consts: list[Constellation] = [
            Constellation(
                id=int(const["id"]),
                name=self._text_map[const["nameTextMapHash"]],
                icon=f"https://enka.network/ui/{const['icon']}.png",
                unlocked=any(const["id"] == str(c.id) for c in c.constellations),
            )
            for const in all_consts
        ]
        c.constellations = consts

    def _post_process_character(self, c: Character) -> None:
        is_no_element_traveler = c.skill_depot_id in {501, 701}
        characer_id = (
            f"{c.id}-{c.skill_depot_id}"
            if c.id in {10000005, 10000007} and not is_no_element_traveler
            else str(c.id)
        )

        data = self._assets.character_data[characer_id]
        # name
        character_name_text_map_hash = data["NameTextMapHash"]
        c.name = self._text_map[character_name_text_map_hash]

        # icon
        side_icon_name = data["SideIconName"]
        c.icon = Icon(side_icon_ui_path=side_icon_name)

        # constellations
        if not is_no_element_traveler:
            self._post_process_constellations(c, data)

        # talents
        self._post_process_talents(c, data)

        # talent order
        c.talent_order = [100540] if is_no_element_traveler else data["SkillOrder"]

        # element
        element = data["Element"]
        c.element = Element.NONE if is_no_element_traveler else Element(element)

        # rarity
        c.rarity = CHARACTER_RARITY_MAP[data["QualityType"]]

        # namecard
        ui_path = data.get("NamecardIcon")
        if ui_path is not None:
            c.namecard = Namecard(ui_path=ui_path)

        # costume
        if c.costume_id is not None:
            costume_data = data.get("Costumes")
            if costume_data is not None:
                c.costume = Costume(id=c.costume_id, data=costume_data[str(c.costume_id)])

        # weapon
        weapon = c.weapon
        self._post_process_weapon(weapon)

        # artifacts
        for artifact in c.artifacts:
            self._post_process_artifact(artifact)

        # stats
        for stat_type, stat in c.stats.items():
            if not isinstance(stat_type, FightPropType):
                continue
            stat.name = self._text_map.get(stat_type.name)

    def _post_process_showcase(self, showcase: ShowcaseResponse) -> None:
        # player
        self._post_process_player(showcase.player)

        # showcase characters
        for character in showcase.player.showcase_characters:
            self._post_process_showcase_character(character)

        # characters
        for character in showcase.characters:
            self._post_process_character(character)

    async def start(self) -> None:
        await super().start()
        await self._assets.load(self.session)
        self._text_map = TextMap(self.lang, self._assets.text_map)

    async def update_assets(self) -> None:
        """Update game assets."""
        logger.info("Updating GI assets")
        await self._assets.update(self.session)
        self._text_map = TextMap(self.lang, self._assets.text_map)
        logger.info("GI assets updated")

    @overload
    async def fetch_showcase(
        self, uid: str | int, *, info_only: bool = False, raw: Literal[False] = False
    ) -> ShowcaseResponse: ...
    @overload
    async def fetch_showcase(
        self, uid: str | int, *, info_only: bool = False, raw: Literal[True] = True
    ) -> dict[str, Any]: ...
    async def fetch_showcase(
        self, uid: str | int, *, info_only: bool = False, raw: bool = False
    ) -> ShowcaseResponse | dict[str, Any]:
        """Fetch the player showcase of the given UID.

        Args:
            uid: The UID of the user.
            info_only: Whether to only fetch player info, defaults to False.
            raw: Whether to return the raw data, defaults to False.

        Returns:
            The parsed or raw showcase data.
        """
        url = API_URL.format(uid=uid)
        if info_only:
            url += "?info"

        data = await self._request(url)
        if raw:
            return data

        data = copy.deepcopy(data)
        showcase = ShowcaseResponse(**data)
        self._post_process_showcase(showcase)
        return showcase

    def parse_showcase(self, data: dict[str, Any]) -> ShowcaseResponse:
        """Parse the given showcase data.

        Args:
            data: The showcase data.

        Returns:
            The parsed showcase response.
        """
        data = copy.deepcopy(data)
        showcase = ShowcaseResponse(**data)
        self._post_process_showcase(showcase)
        return showcase

    async def fetch_builds(self, owner: Owner) -> dict[str, list[Build]]:
        """Fetch the character builds of the given owner.

        Args:
            owner: The owner of the builds.

        Returns:
            Character ID to list of builds mapping.
        """
        url = f"https://enka.network/api/profile/{owner.username}/hoyos/{owner.hash}/builds/"
        data = await self._request(url)
        result: dict[str, list[Build]] = {}

        for key, builds in data.items():
            result[key] = []
            for build in builds:
                build_ = Build(**build)
                self._post_process_character(build_.character)
                result[key].append(build_)

        return result



from __future__ import annotations

import pytest

from enka.gi import Element, GenshinClient, Language


async def test_update_assets(genshin_client: GenshinClient) -> None:
    await genshin_client.update_assets()


async def test_fetch_showcase(genshin_client: GenshinClient) -> None:
    showcase = await genshin_client.fetch_showcase("901211014")
    assert showcase.uid == "901211014"


async def test_empty_showcase(genshin_client: GenshinClient) -> None:
    showcase = await genshin_client.fetch_showcase("123456789")
    assert showcase.uid == "123456789"
    assert len(showcase.characters) == 0


async def test_traveler_showcase(genshin_client: GenshinClient) -> None:
    showcase = await genshin_client.fetch_showcase("600001919")
    assert showcase.uid == "600001919"


async def test_zero_achievement(genshin_client: GenshinClient) -> None:
    showcase = await genshin_client.fetch_showcase("831335714")
    assert showcase.player.achievements == 0


async def test_new_profile_picture_format(genshin_client: GenshinClient) -> None:
    showcase = await genshin_client.fetch_showcase("724824926")
    assert showcase.player.profile_picture_icon is not None


async def test_costume(genshin_client: GenshinClient) -> None:
    showcase = await genshin_client.fetch_showcase("738081787")
    assert showcase.player.showcase_characters[3].costume is not None


@pytest.mark.parametrize("lang", list(Language))
async def test_langs(lang: Language) -> None:
    async with GenshinClient(lang) as api:
        await api.fetch_showcase("901211014")


async def test_owner_and_builds(genshin_client: GenshinClient) -> None:
    showcase = await genshin_client.fetch_showcase("618285856")
    assert showcase.owner is not None
    await genshin_client.fetch_builds(showcase.owner)


async def test_raw_and_parse(genshin_client: GenshinClient) -> None:
    raw = await genshin_client.fetch_showcase("901211014", raw=True)
    genshin_client.parse_showcase(raw)


async def test_invalid_player(genshin_client: GenshinClient) -> None:
    await genshin_client.fetch_showcase("901211015")


async def test_no_element_traveler(genshin_client: GenshinClient) -> None:
    showcase = await genshin_client.fetch_showcase("828107378")
    traveler = next(c for c in showcase.characters if c.name == "Aether")
    assert traveler.element is Element.NONE



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

    import aiohttp

    from .data import AssetData

__all__ = ("AssetManager",)


class AssetManager:
    """Base asset manager."""

    def __init__(self) -> None:
        self._assets: Sequence[AssetData] = ()

    async def load(self, session: aiohttp.ClientSession) -> None:
        tasks: list[asyncio.Task[None]] = [
            asyncio.create_task(asset.load(session)) for asset in self._assets
        ]
        await asyncio.gather(*tasks, return_exceptions=True)

    def clear(self) -> None:
        for asset in self._assets:
            asset._data = None

    async def update(self, session: aiohttp.ClientSession) -> None:
        tasks: list[asyncio.Task[None]] = [
            asyncio.create_task(asset.update(session)) for asset in self._assets
        ]
        await asyncio.gather(*tasks, return_exceptions=True)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from enka.assets.hsr.manager import HSRAssetManager
    from enka.models.hsr import Character, LightCone, Relic

DEFAULT_PROPS: Final[dict[str, float]] = {
    "BaseHP": 0,
    "HPAddedRatio": 0,
    "HPDelta": 0,
    "HPConvert": 0,
    "BaseAttack": 0,
    "AttackAddedRatio": 0,
    "AttackDelta": 0,
    "AttackConvert": 0,
    "BaseDefence": 0,
    "DefenceAddedRatio": 0,
    "DefenceDelta": 0,
    "DefenceConvert": 0,
    "BaseSpeed": 0,
    "SpeedAddedRatio": 0,
    "SpeedDelta": 0,
    "SpeedConvert": 0,
    "CriticalChance": 0,
    "CriticalChanceBase": 0,
    "CriticalDamage": 0,
    "CriticalDamageBase": 0,
    "SPRatio": 0,
    "SPRatioBase": 0,
    "SPRatioConvert": 0,
    "StatusProbability": 0,
    "StatusProbabilityBase": 0,
    "StatusProbabilityConvert": 0,
    "StatusResistance": 0,
    "StatusResistanceBase": 0,
    "StatusResistanceConvert": 0,
    "HealRatioBase": 0,
    "HealRatioConvert": 0,
    "HealTakenRatio": 0,
    "ShieldAddedRatio": 0,
    "ShieldTakenRatio": 0,
    "AggroBase": 0,
    "AggroAddedRatio": 0,
    "AggroDelta": 0,
    "BreakDamageAddedRatio": 0,
    "BreakDamageAddedRatioBase": 0,
    "AllDamageTypeResistance": 0,
    "PhysicalResistanceDelta": 0,
    "FireResistanceDelta": 0,
    "IceResistanceDelta": 0,
    "ThunderResistanceDelta": 0,
    "QuantumResistanceDelta": 0,
    "ImaginaryResistanceDelta": 0,
    "WindResistanceDelta": 0,
    "PhysicalPenetrate": 0,
    "FirePenetrate": 0,
    "IcePenetrate": 0,
    "ThunderPenetrate": 0,
    "QuantumPenetrate": 0,
    "ImaginaryPenetrate": 0,
    "WindPenetrate": 0,
    "AllDamageTypeTakenRatio": 0,
    "PhysicalTakenRatio": 0,
    "FireTakenRatio": 0,
    "IceTakenRatio": 0,
    "ThunderTakenRatio": 0,
    "QuantumTakenRatio": 0,
    "ImaginaryTakenRatio": 0,
    "WindTakenRatio": 0,
    "AllDamageTypeAddedRatio": 0,
    "DOTDamageAddedRatio": 0,
    "PhysicalAddedRatio": 0,
    "FireAddedRatio": 0,
    "IceAddedRatio": 0,
    "ThunderAddedRatio": 0,
    "QuantumAddedRatio": 0,
    "ImaginaryAddedRatio": 0,
    "WindAddedRatio": 0,
    "StanceBreakAddedRatio": 0,
    "AllDamageReduce": 0,
    "FatigueRatio": 0,
    "MinimumFatigueRatio": 0,
}


class LayerGenerator:
    def __init__(self, assets: HSRAssetManager) -> None:
        self._assets = assets

    def character(self, c: Character) -> PropLayer:
        layer = PropLayer()
        promotion = self._assets.meta_data["avatar"][str(c.id)][str(c.ascension)]

        layer.add("BaseHP", promotion["HPBase"] + promotion["HPAdd"] * (c.level - 1))
        layer.add("BaseAttack", promotion["AttackBase"] + promotion["AttackAdd"] * (c.level - 1))
        layer.add("BaseDefence", promotion["DefenceBase"] + promotion["DefenceAdd"] * (c.level - 1))
        layer.add("BaseSpeed", promotion["SpeedBase"])
        layer.add("CriticalChance", promotion["CriticalChance"])
        layer.add("CriticalDamage", promotion["CriticalDamage"])
        layer.add("SPRatio", 1)

        return layer

    def skill_tree(self, c: Character) -> PropLayer:
        layer = PropLayer()
        tree_store = self._assets.meta_data["tree"]

        for trace in c.traces:
            if str(trace.id) not in tree_store:
                continue

            meta = tree_store[str(trace.id)][str(trace.level)]["props"]
            for key, value in meta.items():
                layer.add(key, value)

        return layer

    def light_cone(self, lc: LightCone) -> PropLayer:
        layer = PropLayer()
        promotion = self._assets.meta_data["equipment"][str(lc.id)][str(lc.ascension)]

        base_stats = {
            "BaseHP": promotion["BaseHP"] + promotion["HPAdd"] * (lc.level - 1),
            "BaseAttack": promotion["BaseAttack"] + promotion["AttackAdd"] * (lc.level - 1),
            "BaseDefence": promotion["BaseDefence"] + promotion["DefenceAdd"] * (lc.level - 1),
        }

        for key, value in base_stats.items():
            layer.add(key, value)

        return layer

    def light_cone_skill(self, lc: LightCone) -> PropLayer:
        layer = PropLayer()
        meta = self._assets.meta_data["equipmentSkill"]

        if str(lc.id) in meta:
            meta = meta[str(lc.id)][str(lc.superimpose)]["props"]
            for key, value in meta.items():
                layer.add(key, value)

        return layer

    def relics(self, relics: list[Relic]) -> PropLayer:
        layer = PropLayer()

        for r in relics:
            for stat in r.stats:
                layer.add(stat.type.value, stat.value)

        return layer

    def relic_set(self, relics: list[Relic]) -> PropLayer:
        sets: defaultdict[int, int] = defaultdict(int)
        for r in relics:
            sets[r.set_id] += 1

        layer = PropLayer()
        for relic_set_id, count in sets.items():
            if count >= 2:
                props = self._assets.meta_data["relic"]["setSkill"][str(relic_set_id)]["2"]["props"]
                for key, value in props.items():
                    layer.add(key, value)

            if count >= 4:
                props = self._assets.meta_data["relic"]["setSkill"][str(relic_set_id)]["4"]["props"]
                for key, value in props.items():
                    layer.add(key, value)

        return layer


class PropLayer:
    def __init__(self) -> None:
        self.props = DEFAULT_PROPS.copy()

    def add(self, name: str, value: float) -> None:
        if name not in self.props:
            msg = f"Invalid stat name: {name!r}"
            raise ValueError(msg)

        self.props[name] += value

    @property
    def hp(self) -> float:
        return (
            self.props["BaseHP"] * (1 + self.props["HPAddedRatio"])
            + self.props["HPDelta"]
            + self.props["HPConvert"]
        )

    @property
    def attack(self) -> float:
        return (
            self.props["BaseAttack"] * (1 + self.props["AttackAddedRatio"])
            + self.props["AttackDelta"]
            + self.props["AttackConvert"]
        )

    @property
    def defence(self) -> float:
        return (
            self.props["BaseDefence"] * (1 + self.props["DefenceAddedRatio"])
            + self.props["DefenceDelta"]
            + self.props["DefenceConvert"]
        )

    @property
    def speed(self) -> float:
        return (
            self.props["BaseSpeed"] * (1 + self.props["SpeedAddedRatio"])
            + self.props["SpeedDelta"]
            + self.props["SpeedConvert"]
        )

    @property
    def break_damage(self) -> float:
        return self.props["BreakDamageAddedRatioBase"] + self.props["BreakDamageAddedRatio"]

    @property
    def critical_chance(self) -> float:
        return self.props["CriticalChanceBase"] + self.props["CriticalChance"]

    @property
    def critical_damage(self) -> float:
        return self.props["CriticalDamageBase"] + self.props["CriticalDamage"]

    @property
    def heal_ratio(self) -> float:
        return self.props["HealRatioBase"] + self.props["HealRatioConvert"]

    @property
    def sp_ratio(self) -> float:
        return self.props["SPRatioBase"] + self.props["SPRatio"] + self.props["SPRatioConvert"]

    @property
    def status_probability(self) -> float:
        return self.props["StatusProbabilityBase"] + self.props["StatusProbability"]

    @property
    def status_resistance(self) -> float:
        return self.props["StatusResistanceBase"] + self.props["StatusResistance"]

    @property
    def physical_damage(self) -> float:
        return self.props["PhysicalAddedRatio"] + self.props["AllDamageTypeAddedRatio"]

    @property
    def fire_damage(self) -> float:
        return self.props["FireAddedRatio"] + self.props["AllDamageTypeAddedRatio"]

    @property
    def ice_damage(self) -> float:
        return self.props["IceAddedRatio"] + self.props["AllDamageTypeAddedRatio"]

    @property
    def thunder_damage(self) -> float:
        return self.props["ThunderAddedRatio"] + self.props["AllDamageTypeAddedRatio"]

    @property
    def wind_damage(self) -> float:
        return self.props["WindAddedRatio"] + self.props["AllDamageTypeAddedRatio"]

    @property
    def quantum_damage(self) -> float:
        return self.props["QuantumAddedRatio"] + self.props["AllDamageTypeAddedRatio"]

    @property
    def imaginary_damage(self) -> float:
        return self.props["ImaginaryAddedRatio"] + self.props["AllDamageTypeAddedRatio"]


class PropState:
    def __init__(self) -> None:
        self.layers: list[PropLayer] = []

    def add(self, layer: PropLayer) -> None:
        self.layers.append(layer)

    def sum(self) -> PropLayer:
        sum_layer = PropLayer()
        for layer in self.layers:
            for key, value in layer.props.items():
                sum_layer.add(key, value)

        return sum_layer



from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field, computed_field, field_validator, model_validator

from ...constants.hsr import ASCENSION_TO_MAX_LEVEL, DMG_BONUS_PROPS, PERCENT_STAT_TYPES
from ...enums.hsr import Element, Path, RelicType, StatType, TraceType
from ...utils import round_down
from .icon import CharacterIcon, LightConeIcon

__all__ = ("Character", "LightCone", "Relic", "RelicSubAffix", "Stat", "Trace")


class Trace(BaseModel):
    """Represents a character's trace (skill).

    Attributes:
        id: The trace's ID.
        level: The trace's level.
        boosted: Whether the trace's level is boosted by an activated eidolon's effect.
        icon: The trace's icon.
        max_level: The trace's maximum level.
        anchor: The trace's anchor.
        type: The trace's type.
    """

    id: int = Field(alias="pointId")
    level: int
    boosted: bool = False
    """Whether the level of this trace is boosted by an activated eidolon's effect."""

    # Following fields are added in post-processing
    icon: str = ""
    max_level: int = 0
    anchor: str = ""
    type: TraceType = TraceType.UNKNOWN


class Stat(BaseModel):
    """Represents a HSR stat.

    Attributes:
        type: The type of the stat.
        value: The value of the stat.
        name: The name of the stat.
        icon: The icon of the stat.
    """

    type: StatType
    value: float

    # Following fields are added in post-processing
    name: str = ""
    icon: str = ""

    @computed_field
    @property
    def is_percentage(self) -> bool:
        """Whether the stat is a percentage stat."""
        return self.type.value in PERCENT_STAT_TYPES

    @computed_field
    @property
    def formatted_value(self) -> str:
        """The formatted value of the stat."""
        if self.is_percentage:
            return f"{round_down(self.value * 100, 1)}%"
        if self.type in {StatType.SPD, StatType.SPEED_DELTA}:
            return str(round_down(self.value, 2))
        return str(int(round_down(self.value, 0)))


class LightCone(BaseModel):
    """Represents a light cone (weapon.)

    Attributes:
        id: The light cone's ID.
        level: The light cone's level.
        ascension: The light cone's ascension level.
        superimpose: The light cone's superimpose level.
        name: The name of the light cone.
        stats: The stats of the light cone.
        rarity: The rarity of the light cone.
        icon: The icon of the light cone.
        path: The light cone's path.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: int = Field(alias="tid")
    level: int
    ascension: Literal[0, 1, 2, 3, 4, 5, 6] = Field(0, alias="promotion")
    superimpose: Literal[1, 2, 3, 4, 5] = Field(alias="rank")
    name: str  # Returned as text map hash in the API response
    stats: list[Stat] = Field(alias="props")

    # Following fields are added in post-processing
    icon: LightConeIcon = LightConeIcon(light_cone_id=0)
    rarity: Literal[3, 4, 5] = 3
    path: Path = Path.NONE

    @computed_field
    @property
    def max_level(self) -> Literal[20, 30, 40, 50, 60, 70, 80]:
        """Light Cone's max level."""
        return ASCENSION_TO_MAX_LEVEL[self.ascension]

    @field_validator("name", mode="before")
    @classmethod
    def _stringify_name(cls, value: int) -> str:
        return str(value)

    @model_validator(mode="before")
    @classmethod
    def _flatten_flat(cls, values: dict[str, Any]) -> dict[str, Any]:
        flat_ = values.pop("_flat")
        values.update(flat_)
        return values


class RelicSubAffix(BaseModel):
    """Represents a relic's sub-stat information.

    Attributes:
        id: The ID of the sub-stat.
        cnt: The count of the sub-stat.
        step: The step of the sub-stat.
    """

    id: int = Field(alias="affixId")
    cnt: int
    step: int | None = None


class Relic(BaseModel):
    """Represents a relic in HSR.

    Attributes:
        id: The relic's ID.
        level: The relic's level.
        type: The relic's type.
        main_affix_id: The ID of the main affix.
        set_name: The name of the relic set.
        set_id: The ID of the relic set.
        stats: The stats of the relic.
        sub_affix_list: List of sub-stat information for the relic.
        icon: The icon of the relic.
        rarity: The rarity of the relic.
    """

    id: int = Field(alias="tid")
    level: int = 0
    type: RelicType
    main_affix_id: int = Field(alias="mainAffixId")
    set_name: str = Field(alias="setName")  # Returned as text map hash in the API response
    set_id: int = Field(alias="setID")
    stats: list[Stat] = Field(alias="props")
    sub_affix_list: list[RelicSubAffix] = Field(alias="subAffixList", default_factory=list)

    # The following fields are added in post-processing
    icon: str = ""
    rarity: Literal[3, 4, 5] = 3

    @field_validator("set_name", mode="before")
    @classmethod
    def _stringify_set_name(cls, value: int) -> str:
        return str(value)

    @model_validator(mode="before")
    @classmethod
    def _flatten_flat(cls, values: dict[str, Any]) -> dict[str, Any]:
        flat_ = values.pop("_flat")
        values.update(flat_)
        return values

    @computed_field
    @property
    def main_stat(self) -> Stat:
        """The relic's main stat."""
        return self.stats[0]

    @computed_field
    @property
    def sub_stats(self) -> list[Stat]:
        """The relic's sub-stats."""
        return self.stats[1:]


class Eidolon(BaseModel):
    """Represents a character's eidolon.

    Attributes:
        id: The eidolon's ID.
        icon: The eidolon's icon.
        unlocked: Whether the eidolon is unlocked.
    """

    id: int
    icon: str
    unlocked: bool


class Character(BaseModel):
    """Represents a character in HSR.

    Attributes:
        level: The character's level.
        ascension: The character's ascension level.
        id: The character's ID.
        traces: List of the character's traces (skills).
        light_cone: The character's light cone (weapon).
        relics: List of the character's relics.
        eidolons: List of the character's eidolons.
        eidolons_unlocked: The number of unlocked eidolons.
        is_assist: Whether the character is an assist character.
        icon: The character's icon.
        name: The character's name.
        rarity: The character's rarity.
        element: The character's element (type.)
        path: The character's path.
        stats: The character's stats.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    level: int
    ascension: Literal[0, 1, 2, 3, 4, 5, 6] = Field(0, alias="promotion")
    id: int = Field(alias="avatarId")
    traces: list[Trace] = Field(alias="skillTreeList")
    light_cone: LightCone | None = Field(None, alias="equipment")
    relics: list[Relic] = Field(alias="relicList", default_factory=list)
    eidolons: list[Eidolon] = Field(default_factory=list)
    eidolons_unlocked: int = Field(0, alias="rank")
    is_assist: bool = Field(False, alias="_assist")

    # Following fields are added in post-processing
    icon: CharacterIcon = CharacterIcon(character_id=0)
    name: str = ""
    rarity: Literal[4, 5] = 4
    element: Element = Element.FIRE
    path: Path = Path.ABUNDANCE
    stats: dict[StatType, Stat] = Field(default_factory=dict)

    @computed_field
    @property
    def max_level(self) -> Literal[20, 30, 40, 50, 60, 70, 80]:
        """Character's max level."""
        return ASCENSION_TO_MAX_LEVEL[self.ascension]

    @computed_field
    @property
    def highest_dmg_bonus_stat(self) -> Stat:
        """Character's highest damage bonus stat."""
        return max(
            (stat for stat in self.stats.values() if stat.type in DMG_BONUS_PROPS.values()),
            key=lambda stat: stat.value,
            default=next(
                stat
                for stat in self.stats.values()
                if stat.type.value == DMG_BONUS_PROPS[self.element]
            ),
        )



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field, field_validator

from ..enka.owner import Owner
from .character import Character
from .player import Player

__all__ = ("ShowcaseResponse",)


class ShowcaseResponse(BaseModel):
    """Represents a Genshin Impact showcase response.

    Attributes:
        characters: The characters in the showcase.
        player: The player of the showcase.
        ttl: The time to live of the response.
        uid: The UID of the showcase.
        owner: The owner of the showcase's account.
    """

    characters: list[Character] = Field(alias="avatarInfoList", default_factory=list)
    player: Player = Field(alias="playerInfo")
    ttl: int
    uid: str
    owner: Owner | None = None

    @field_validator("characters", mode="before")
    @classmethod
    def __handle_none_value(cls, v: list[dict[str, Any]] | None) -> list[dict[str, Any]]:
        if v is None:
            return []
        return v



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field, field_validator, model_validator

from .character import Agent
from .player import Player

__all__ = ("ShowcaseResponse",)


class ShowcaseResponse(BaseModel):
    """Represents a ZZZ player's showcase information.

    Attributes:
        agents: List of agents (characters) in the player's showcase.
        player: The player's profile information.
        uid: The player's unique identifier.
        ttl: The time-to-live for the showcase data.
    """

    agents: list[Agent] = Field(alias="ShowcaseDetail")
    player: Player = Field(alias="SocialDetail")

    uid: int
    ttl: int

    @field_validator("agents", mode="before")
    @classmethod
    def __get_agents(cls, value: dict[str, Any]) -> list[dict[str, Any]]:
        return value["AvatarList"]

    @model_validator(mode="before")
    @classmethod
    def __unnest_info(cls, value: dict[str, Any]) -> dict[str, Any]:
        info = value.pop("PlayerInfo")
        value.update(info)
        return value

    @property
    def url(self) -> str:
        """The URL of the showcase."""
        return f"https://enka.network/zzz/{self.uid}"
